""" Adapted from Daniel Machado's REFRAMED"""
from .solver import Solver, VarType, Parameter, default_parameters
from .solution import Solution, Status
from cplex import Cplex, infinity, SparsePair
import sys
from math import inf
from warnings import warn


def infinity_fix(val):
    if val == inf:
        return infinity
    elif val == -inf:
        return -infinity
    else:
        return val


class CplexSolver(Solver):
    """ Implements the solver interface using CPLEX. """

    def __init__(self, model=None):
        Solver.__init__(self)
        self.problem = Cplex()

        self.status_mapping = {
            self.problem.solution.status.optimal: Status.OPTIMAL,
            self.problem.solution.status.optimal_tolerance: Status.OPTIMAL,
            self.problem.solution.status.unbounded: Status.UNBOUNDED,
            self.problem.solution.status.infeasible: Status.INFEASIBLE,
            self.problem.solution.status.infeasible_or_unbounded: Status.INF_OR_UNB,
            self.problem.solution.status.MIP_optimal: Status.OPTIMAL,
            self.problem.solution.status.MIP_unbounded: Status.UNBOUNDED,
            self.problem.solution.status.MIP_infeasible: Status.INFEASIBLE,
            self.problem.solution.status.MIP_infeasible_or_unbounded: Status.INF_OR_UNB
        }

        self.vartype_mapping = {
            VarType.BINARY: self.problem.variables.type.binary,
            VarType.INTEGER: self.problem.variables.type.integer,
            VarType.CONTINUOUS: self.problem.variables.type.continuous
        }

        self.parameter_mapping = {
            Parameter.TIME_LIMIT: self.problem.parameters.timelimit,
            Parameter.FEASIBILITY_TOL: self.problem.parameters.simplex.tolerances.feasibility,
            Parameter.OPTIMALITY_TOL: self.problem.parameters.simplex.tolerances.optimality,
            Parameter.INT_FEASIBILITY_TOL: self.problem.parameters.mip.tolerances.integrality,
            Parameter.MIP_ABS_GAP: self.problem.parameters.mip.tolerances.mipgap,
            Parameter.MIP_REL_GAP: self.problem.parameters.mip.tolerances.absmipgap,
            Parameter.POOL_SIZE: self.problem.parameters.mip.limits.populate,
            Parameter.POOL_GAP: self.problem.parameters.mip.pool.relgap
        }

        self.set_parameters(default_parameters)

        self.set_logging(False)
        self._cached_lin_obj = {}
        self._cached_sense = None
        self._cached_lower_bounds = {}
        self._cached_upper_bounds = {}
        self._cached_vars = []
        self._cached_constrs = []

        if model:
            self.build_problem(model)

    def add_variable(self, var_id, lb=-inf, ub=inf, vartype=VarType.CONTINUOUS, update=True):
        """ Add a variable to the current problem.

        Arguments:
            var_id (str): variable identifier
            lb (float): lower bound
            ub (float): upper bound
            vartype (VarType): variable type (default: CONTINUOUS)
            update (bool): update problem immediately
        """

        if update:
            self.add_variables([var_id], [lb], [ub], [vartype])
        else:
            self._cached_vars.append((var_id, lb, ub, vartype))

    def add_variables(self, var_ids, lbs, ubs, vartypes):
        """ Add multiple variables to the current problem.

        Arguments:
            var_ids (list): variable identifier
            lbs (list): lower bounds
            ubs (list): upper bounds
            vartypes (list): variable types (default: CONTINUOUS)
        """

        lbs = list(map(infinity_fix, lbs))
        ubs = list(map(infinity_fix, ubs))

        if set(vartypes) == {VarType.CONTINUOUS}:
            self.problem.variables.add(names=var_ids, lb=lbs, ub=ubs)
        else:
            vartypes = [self.vartype_mapping[vartype] for vartype in vartypes]
            self.problem.variables.add(names=var_ids, lb=lbs, ub=ubs, types=vartypes)

        self.var_ids.extend(var_ids)
        self._cached_lower_bounds.update(dict(zip(var_ids, lbs)))
        self._cached_upper_bounds.update(dict(zip(var_ids, ubs)))
        self._cached_lin_obj.update({var_id: 0.0 for var_id in var_ids})

    def set_variable_bounds(self, var_id, lb, ub):
        """Modify a variable bounds

        Args:
            var_id (str): [description]
            lb (float): lower bound
            ub (float): upper bound
        """
        if lb:
            self.problem.variables.set_lower_bounds(var_id, lb)
        if ub:
            self.problem.variables.set_upper_bounds(var_id, ub)

    def add_constraint(self, constr_id, lhs, sense='=', rhs=0, update=True):
        """ Add a constraint to the current problem.

        Arguments:
            constr_id (str): constraint identifier
            lhs (dict): variables and respective coefficients
            sense (str): constraint sense (any of: '<', '=', '>'; default '=')
            rhs (float): right-hand side of equation (default: 0)
            update (bool): update problem immediately
        """

        if update:
            self.add_constraints([constr_id], [lhs], [sense], [rhs])
        else:
            self._cached_constrs.append((constr_id, lhs, sense, rhs))

    def add_constraints(self, constr_ids, lhs, senses, rhs):
        """ Add a list of constraints to the current problem.

        Arguments:
            constr_ids (list): constraint identifiers
            lhs (list): variables and respective coefficients
            senses (list): constraint senses (default: '=')
            rhs (list): right-hand side of equations (default: 0)
        """

        map_sense = {'=': 'E',
                     '<': 'L',
                     '>': 'G'}

        exprs = [SparsePair(ind=list(constr.keys()), val=list(constr.values())) for constr in lhs]
        senses = [map_sense[sense] for sense in senses]

        self.problem.linear_constraints.add(lin_expr=exprs,
                                            senses=senses,
                                            rhs=rhs,
                                            names=constr_ids)

        self.constr_ids.extend(constr_ids)

    def remove_variable(self, var_id):
        """ Remove a variable from the current problem.

        Arguments:
            var_id (str): variable identifier
        """
        self.remove_variables([var_id])

    def remove_variables(self, var_ids):
        """ Remove variables from the current problem.

        Arguments:
            var_ids (list): variable identifiers
        """

        found = []
        for var_id in var_ids:
            if var_id in self.var_ids:
                found.append(var_id)
                self.var_ids.remove(var_id)

        self.problem.variables.delete(found)

    def remove_constraint(self, constr_id):
        """ Remove a constraint from the current problem.

        Arguments:
            constr_id (str): constraint identifier
        """
        self.remove_constraints([constr_id])

    def remove_constraints(self, constr_ids):
        """ Remove constraints from the current problem.

        Arguments:
            constr_ids (list): constraint identifiers
        """

        found = []
        for constr_id in constr_ids:
            if constr_id in self.constr_ids:
                found.append(constr_id)
                self.constr_ids.remove(constr_id)

        self.problem.linear_constraints.delete(found)

    def update(self):
        """ Update internal structure. Used for efficient lazy updating. """

        if self._cached_vars:
            var_ids = [x[0] for x in self._cached_vars]
            lbs = [x[1] for x in self._cached_vars]
            ubs = [x[2] for x in self._cached_vars]
            vartypes = [x[3] for x in self._cached_vars]
            self.add_variables(var_ids, lbs, ubs, vartypes)
            self._cached_vars = []

        if self._cached_constrs:
            constr_ids = [x[0] for x in self._cached_constrs]
            lhs = [x[1] for x in self._cached_constrs]
            senses = [x[2] for x in self._cached_constrs]
            rhs = [x[3] for x in self._cached_constrs]
            self.add_constraints(constr_ids, lhs, senses, rhs)
            self._cached_constrs = []

    def set_objective(self, linear=None, quadratic=None, minimize=True):
        """ Set a predefined objective for this problem.

        Args:
            linear (str or dict): linear coefficients (or a single variable to optimize)
            quadratic (dict): quadratic coefficients (optional)
            minimize (bool): solve a minimization problem (default: True)

        Notes:
            Setting the objective is optional. It can also be passed directly when calling **solve**.

        """

        if linear:

            if isinstance(linear, str):
                linear = {linear: 1.0}

            updated_coeffs = {}

            for var_id in self.var_ids:
                if var_id in linear and linear[var_id] != self._cached_lin_obj[var_id]:
                    updated_coeffs[var_id] = linear[var_id]
                if var_id not in linear and self._cached_lin_obj[var_id] != 0.0:
                    updated_coeffs[var_id] = 0.0

            if updated_coeffs:
                self.problem.objective.set_linear(list(updated_coeffs.items()))
                self._cached_lin_obj.update(updated_coeffs)

            for r_id in linear:
                if r_id not in self.var_ids:
                    warn(f"Objective variable not previously declared: {r_id}")

        if quadratic:
            self.problem.objective.set_quadratic([0.0] * len(self.var_ids))
            quad_coeffs = [(r_id1, r_id2, coeff) for (r_id1, r_id2), coeff in quadratic.items()]
            self.problem.objective.set_quadratic_coefficients(quad_coeffs)

            for (r_id1, r_id2) in quadratic:
                if r_id1 not in self.var_ids:
                    warn(f"Objective variable not previously declared: {r_id1}")
                if r_id2 not in self.var_ids:
                    warn(f"Objective variable not previously declared: {r_id2}")

        if minimize != self._cached_sense:
            if minimize:
                sense = self.problem.objective.sense.minimize
            else:
                sense = self.problem.objective.sense.maximize
            self.problem.objective.set_sense(sense)
            self._cached_sense = minimize

    def __build_problem_simulator(self, model):
        """ Create problem structure for a given model.

        Arguments:
            model : Simulator
        """
        var_ids = []
        lbs = []
        ubs = []
        for r_id in model.reactions:
            lb, ub = model.get_reaction_bounds(r_id)
            var_ids.append(r_id)
            lbs.append(lb)
            ubs.append(ub)

        var_types = [VarType.CONTINUOUS] * len(var_ids)
        self.add_variables(var_ids, lbs, ubs, var_types)

        constr_ids = list(model.metabolites.keys())
        table = model.metabolite_reaction_lookup()
        lhs = list(table.values())
        senses = ['='] * len(constr_ids)
        rhs = [0] * len(constr_ids)
        self.add_constraints(constr_ids, lhs, senses, rhs)

    def solve(self, linear=None, quadratic=None, minimize=None, model=None, constraints=None, get_values=True,
              shadow_prices=False, reduced_costs=False, pool_size=0, pool_gap=None):
        """ Solve the optimization problem.

        Arguments:
            linear (str or dict): linear coefficients (or a single variable to optimize)
            quadratic (dict): quadratic objective (optional)
            minimize (bool): solve a minimization problem (default: True)
            model (CBModel): model (optional, leave blank to reuse previous model structure)
            constraints (dict): additional constraints (optional)
            get_values (bool or list): set to false for speedup if you only care about the objective\
                value (default: True)
            shadow_prices (bool): return shadow prices if available (default: False)
            reduced_costs (bool): return reduced costs if available (default: False)
            pool_size (int): calculate solution pool of given size (only for MILP problems)
            pool_gap (float): maximum relative gap for solutions in pool (optional)

        Returns:
            Solution: solution
        """

        if model:
            self.build_problem(model)

        problem = self.problem

        if constraints:
            changed_lb, changed_ub = self.temporary_bounds(constraints)

        self.set_objective(linear, quadratic, minimize)

        # run the optimization

        if pool_size <= 1:

            problem.solve()

            status = self.status_mapping.get(problem.solution.get_status(), Status.UNKNOWN)
            message = str(problem.solution.get_status_string())

            if status == Status.OPTIMAL:
                fobj = problem.solution.get_objective_value()
                values, s_prices, r_costs = None, None, None

                if get_values:
                    try:
                        get_values = list(get_values)
                        values = dict(zip(get_values, problem.solution.get_values(get_values)))
                    except Exception:
                        values = dict(zip(self.var_ids, problem.solution.get_values()))

                if shadow_prices:
                    s_prices = dict(zip(self.constr_ids, problem.solution.get_dual_values(self.constr_ids)))

                if reduced_costs:
                    r_costs = dict(zip(self.var_ids, problem.solution.get_reduced_costs(self.var_ids)))

                solution = Solution(status, message, fobj, values, s_prices, r_costs)
            else:
                solution = Solution(status, message)

        else:
            pool_pmap = {
                'SolnPoolIntensity': problem.parameters.mip.pool.intensity,
                'PopulateLim': problem.parameters.mip.limits.populate,
                'SolnPoolCapacity': problem.parameters.mip.pool.capacity,
                'SolnPoolReplace': problem.parameters.mip.pool.replace,
                'SolnPoolGap': problem.parameters.mip.pool.relgap,
                'SolnPoolAGap': problem.parameters.mip.pool.absgap

            }

            default_params = {
                'SolnPoolIntensity': 3,
                'PopulateLim': 10 * pool_size,
                'SolnPoolCapacity': pool_size,
                'SolnPoolReplace': 1
            }

            for param, val in default_params.items():
                pool_pmap[param].set(val)

            if pool_gap:
                pool_pmap['SolnPoolGap'].set(pool_gap)

            problem.populate_solution_pool()

            status = self.status_mapping.get(problem.solution.get_status(), Status.UNKNOWN)

            if status == Status.OPTIMAL or status == Status.UNKNOWN:
                solution = self.get_solution_pool(get_values)
            else:
                solution = []

        if constraints:
            self.reset_bounds(changed_lb, changed_ub)

        return solution

    def temporary_bounds(self, constraints):

        lower_bounds, upper_bounds = {}, {}
        lb_new, ub_new = {}, {}

        def _dict_diff(dict1, dict2):
            return set(dict1.items()) - set(dict2.items())

        for r_id, x in constraints.items():
            if r_id in self.var_ids:
                lb, ub = x if isinstance(x, tuple) else (x, x)
                lower_bounds[r_id] = infinity_fix(lb)
                upper_bounds[r_id] = infinity_fix(ub)
            else:
                warn(f"Constrained variable not previously declared: {r_id}")

        if lower_bounds != self._cached_lower_bounds:
            lb_new = _dict_diff(lower_bounds, self._cached_lower_bounds)
            if len(lb_new) > 0:
                self.problem.variables.set_lower_bounds(lb_new)

        if upper_bounds != self._cached_upper_bounds:
            ub_new = _dict_diff(upper_bounds, self._cached_upper_bounds)
            if len(ub_new) > 0:
                self.problem.variables.set_upper_bounds(ub_new)

        return lb_new, ub_new

    def get_solution_pool(self, get_values=True):
        """ Return a solution pool for MILP problems.
        Must be called after using solve with pool_size argument > 0.

        Arguments:
            get_values (bool or list): set to false for speedup if you only care about the\
                objective value (default: True)

        Returns:
            list: list of Solution objects

        """
        pool = self.problem.solution.pool
        solutions = []

        for i in range(pool.get_num()):
            obj = pool.get_objective_value(i)

            if get_values:
                try:
                    get_values = list(get_values)
                    values = dict(zip(get_values, pool.get_values(i, get_values)))
                except Exception:
                    values = dict(zip(self.var_ids, pool.get_values(i)))
            else:
                values = None
            sol = Solution(fobj=obj, values=values)
            solutions.append(sol)

        return solutions

    def reset_bounds(self, updated_lb, updated_ub):
        if updated_lb:
            lb_old = [(r_id, self._cached_lower_bounds[r_id]) for r_id, _ in updated_lb]
            self.problem.variables.set_lower_bounds(lb_old)
        if updated_ub:
            ub_old = [(r_id, self._cached_upper_bounds[r_id]) for r_id, _ in updated_ub]
            self.problem.variables.set_upper_bounds(ub_old)

    def set_parameter(self, parameter, value):
        """ Set a parameter value for this optimization problem

        Arguments:
            parameter (Parameter): parameter type
            value (float): parameter value
        """

        if parameter in self.parameter_mapping:
            self.parameter_mapping[parameter].set(value)
        else:
            raise RuntimeError('Parameter unknown (or not yet supported).')

    def set_logging(self, enabled=False):
        """ Enable or disable log output:

        Arguments:
            enabled (bool): turn logging on (default: False)
        """

        if enabled:
            self.problem.set_log_stream(sys.stdout)
            self.problem.set_error_stream(sys.stderr)
            self.problem.set_warning_stream(sys.stderr)
            self.problem.set_results_stream(sys.stdout)
        else:
            self.problem.set_log_stream(None)
            self.problem.set_error_stream(None)
            self.problem.set_warning_stream(None)
            self.problem.set_results_stream(None)

    def write_to_file(self, filename):
        """ Write problem to file:

        Arguments:
            filename (str): file path
        """

        self.problem.write(filename)



""" Adapted from Daniel Machado's REFRAMED"""
from optlang import Model, Variable, Constraint, Objective
from optlang.symbolics import Zero, add
from .solver import Solver, VarType, Parameter, default_parameters
from .solution import Solution, Status
from math import inf
from warnings import warn


status_mapping = {
    'optimal': Status.OPTIMAL,
    'unbounded': Status.UNBOUNDED,
    'infeasible': Status.INFEASIBLE,
    'infeasible_or_unbounded': Status.INF_OR_UNB,
    'suboptimal': Status.SUBOPTIMAL,
}


class OptLangSolver(Solver):
    """ Implements the gurobi solver interface. """

    def __init__(self, model=None):
        Solver.__init__(self)
        self.problem = Model()

        self.parameter_mapping = {
            Parameter.TIME_LIMIT: self.problem.configuration.timeout,
            Parameter.FEASIBILITY_TOL: self.problem.configuration.tolerances.feasibility,
            Parameter.OPTIMALITY_TOL: self.problem.configuration.tolerances.optimality,
            Parameter.INT_FEASIBILITY_TOL: self.problem.configuration.tolerances.integrality,
        }

        self.set_parameters(default_parameters)
        self.set_logging(False)

        if model:
            self.build_problem(model)

    def add_variable(self, var_id, lb=-inf, ub=inf, vartype=VarType.CONTINUOUS, update=True):
        """ Add a variable to the current problem.

        Arguments:
            var_id (str): variable identifier
            lb (float): lower bound
            ub (float): upper bound
            vartype (VarType): variable type (default: CONTINUOUS)
            update (bool): update problem immediately (default: True)
        """

        if var_id in self.var_ids:
            var = self.problem.variables[var_id]
            var.lb = lb
            var.ub = ub
            var.type = vartype.value
        else:
            var = Variable(var_id, lb=lb, ub=ub, type=vartype.value)
            self.problem.add(var)
            self.var_ids.append(var_id)

        if update:
            self.problem.update()

    def set_variable_bounds(self, var_id, lb, ub):
        """Modify a variable bounds

        Args:
            var_id (str): [description]
            lb (float): lower bound
            ub (float): upper bound
        """
        var = self.problem.variables[var_id]
        if lb:
            var.lb = lb
        if ub:
            var.ub = ub

    def add_constraint(self, constr_id, lhs, sense='=', rhs=0, update=True):
        """ Add a constraint to the current problem.

        Arguments:
            constr_id (str): constraint identifier
            lhs (dict): variables and respective coefficients
            sense (str): constraint sense (any of: '<', '=', '>'; default '=')
            rhs (float): right-hand side of equation (default: 0)
            update (bool): update problem immediately (default: True)
        """

        if constr_id in self.constr_ids:
            self.problem.remove(constr_id)

        if sense == '=':
            constr = Constraint(Zero, lb=rhs, ub=rhs, name=constr_id)
        elif sense == '>':
            constr = Constraint(Zero, lb=rhs, name=constr_id)
        elif sense == '<':
            constr = Constraint(Zero, ub=rhs, name=constr_id)
        else:
            raise RuntimeError(f"Invalid constraint direction: {sense}")

        self.problem.add(constr)
        self.constr_ids.append(constr_id)

        expr = {self.problem.variables[r_id]: coeff for r_id, coeff in lhs.items() if coeff}
        self.problem.constraints[constr_id].set_linear_coefficients(expr)

        if update:
            self.problem.update()

    def remove_variable(self, var_id):
        """ Remove a variable from the current problem.

        Arguments:
            var_id (str): variable identifier
        """
        self.remove_variables([var_id])

    def remove_variables(self, var_ids):
        """ Remove variables from the current problem.

        Arguments:
            var_ids (list): variable identifiers
        """

        for var_id in var_ids:
            if var_id in self.var_ids:
                self.problem.remove(var_id)
                self.var_ids.remove(var_id)

    def remove_constraint(self, constr_id):
        """ Remove a constraint from the current problem.

        Arguments:
            constr_id (str): constraint identifier
        """
        self.remove_constraints([constr_id])

    def remove_constraints(self, constr_ids):
        """ Remove constraints from the current problem.

        Arguments:
            constr_ids (list): constraint identifiers
        """

        for constr_id in constr_ids:
            if constr_id in self.constr_ids:
                self.problem.remove(constr_id)
                self.constr_ids.remove(constr_id)

    def set_objective(self, linear=None, quadratic=None, minimize=True):
        """ Set a predefined objective for this problem.

        Args:
            linear (dict): linear coefficients (optional)
            quadratic (dict): quadratic coefficients (optional)
            minimize (bool): solve a minimization problem (default: True)

        Notes:
            Setting the objective is optional. It can also be passed directly when calling **solve**.

        """

        if linear is None:
            linear = {}

        if quadratic is None:
            quadratic = {}

        if linear and not quadratic:
            objective = {}

            if isinstance(linear, str):
                objective = {self.problem.variables[linear]: 1}
                if linear not in self.var_ids:
                    warn(f"Objective variable not previously declared: {linear}")
            else:
                for r_id, val in linear.items():
                    if r_id not in self.var_ids:
                        warn(f"Objective variable not previously declared: {r_id}")
                    elif val != 0:
                        objective[self.problem.variables[r_id]] = val

            self.problem.objective = Objective(Zero, direction=('min' if minimize else 'max'), sloppy=True)
            self.problem.objective.set_linear_coefficients(objective)
        else:
            objective = []

            for r_id, val in linear.items():
                if r_id not in self.var_ids:
                    warn(f"Objective variable not previously declared: {r_id}")
                elif val != 0:
                    objective.append(val * self.problem.variables[r_id])

            for (r_id1, r_id2), val in quadratic.items():
                if r_id1 not in self.var_ids:
                    warn(f"Objective variable not previously declared: {r_id1}")
                elif r_id2 not in self.var_ids:
                    warn(f"Objective variable not previously declared: {r_id2}")
                elif val != 0:
                    objective.append(val * self.problem.variables[r_id1] * self.problem.variables[r_id2])

            objective_expr = add(objective)
            self.problem.objective = Objective(objective_expr, direction=('min' if minimize else 'max'), sloppy=True)

    def solve(self, linear=None, quadratic=None, minimize=None, model=None, constraints=None, get_values=True,
              shadow_prices=False, reduced_costs=False, pool_size=0, pool_gap=None):
        """ Solve the optimization problem.

        Arguments:
            linear (str or dict): linear coefficients (or a single variable to optimize)
            quadratic (dict): quadratic objective (optional)
            minimize (bool): solve a minimization problem (default: True)
            model (CBModel): model (optional, leave blank to reuse previous model structure)
            constraints (dict): additional constraints (optional)
            get_values (bool or list): set to false for speedup if you only care about the objective (default: True)
            shadow_prices (bool): return shadow prices if available (default: False)
            reduced_costs (bool): return reduced costs if available (default: False)
            pool_size (int): calculate solution pool of given size (only for MILP problems)
            pool_gap (float): maximum relative gap for solutions in pool (optional)

        Returns:
            Solution: solution
        """

        if model:
            self.build_problem(model)

        problem = self.problem

        if constraints:
            old_constraints = {}
            for r_id, x in constraints.items():
                lb, ub = x if isinstance(x, tuple) else (x, x)
                if r_id in self.var_ids:
                    lpvar = problem.variables[r_id]
                    old_constraints[r_id] = (lpvar.lb, lpvar.ub)
                    lpvar.lb, lpvar.ub = lb, ub
                else:
                    warn(f"Constrained variable '{r_id}' not previously declared")
            problem.update()

        self.set_objective(linear, quadratic, minimize)

        # run the optimization
        if pool_size > 1:
            raise RuntimeError("OptLang interface does not support solution pools.")

        problem.optimize()

        status = status_mapping.get(problem.status, Status.UNKNOWN)
        message = str(problem.status)

        if status == Status.OPTIMAL:
            fobj = problem.objective.value
            values, s_prices, r_costs = None, None, None

            if get_values:
                values = dict(problem.primal_values)

                try:
                    values = {x: values[x] for x in get_values}
                except Exception:
                    pass

            if shadow_prices:
                s_prices = dict(problem.shadow_prices)

            if reduced_costs:
                r_costs = dict(problem.reduced_costs)

            solution = Solution(status, message, fobj, values, s_prices, r_costs)
        else:
            solution = Solution(status, message)

        # restore values of temporary constraints
        if constraints:
            for r_id, (lb, ub) in old_constraints.items():
                lpvar = problem.variables[r_id]
                lpvar.lb, lpvar.ub = lb, ub
            problem.update()

        return solution

    def set_parameter(self, parameter, value):
        """ Set a parameter value for this optimization problem

        Arguments:
            parameter (Parameter): parameter type
            value (float): parameter value
        """

        if parameter in self.parameter_mapping:
            self.parameter_mapping[parameter] = value
        else:
            raise RuntimeError('Parameter unknown (or not yet supported).')

    def set_logging(self, enabled=False):
        """ Enable or disable log output:

        Arguments:
            enabled (bool): turn logging on (default: False)
        """

        self.problem.configuration.verbosity = 3 if enabled else 0

    def write_to_file(self, filename):
        """ Write problem to file:

        Arguments:
            filename (str): file path
        """

        with open(filename, "w") as f:
            f.write(self.problem.to_lp())



"""
Simulation for COBRApy models
"""
import logging
from collections import OrderedDict

import numpy as np
from cobra.core.model import Model
from cobra.core.solution import Solution
from cobra.flux_analysis import pfba, moma, room

from . import get_default_solver, SimulationMethod, SStatus
from .simulation import Simulator, SimulationResult, ModelContainer
from ..util.constants import ModelConstants
from ..util.parsing import evaluate_expression_tree
from tqdm import tqdm


LOGGER = logging.getLogger(__name__)


class CobraModelContainer(ModelContainer):
    """ A basic container for COBRApy models.

    :param model: A metabolic model.

    """

    def __init__(self, model: Model = None):
        self.model = model

    @property
    def id(self):
        return self.model.id

    @property
    def reactions(self):
        return [rxn.id for rxn in self.model.reactions]

    def get_reaction(self, r_id):
        rxn = self.model.reactions.get_by_id(r_id)
        stoichiometry = {met.id: val for met, val in rxn.metabolites.items()}
        res = {'id': r_id, 'name': rxn.name, 'lb': rxn.lower_bound,
               'ub': rxn.upper_bound, 'stoichiometry': stoichiometry}
        res['gpr'] = rxn.gene_reaction_rule if rxn.gene_reaction_rule is not None else None
        return res

    @property
    def genes(self):
        return [gene.id for gene in self.model.genes]

    def get_gene(self, g_id):
        g = self.model.genes.get_by_id(g_id)
        res = {'id': g_id, 'name': g.name}
        return res

    @property
    def metabolites(self):
        return [met.id for met in self.model.metabolites]

    def get_metabolite(self, m_id):
        met = self.model.metabolites.get_by_id(m_id)
        res = {'id': m_id, 'name': met.name, 'compartment': met.compartment, 'formula': met.formula}
        return res

    @property
    def medium(self):
        return self.model.medium

    @property
    def compartments(self):
        return self.model.compartments

    def get_compartment(self, c_id):
        c = self.model.compartments[c_id]
        from cobra.medium import find_external_compartment
        e = find_external_compartment(self.model)
        res = {'id': c_id, 'name': c, 'external': (e == c_id)}
        return res

    def get_gpr(self, reaction_id):
        """Returns the gpr rule (str) for a given reaction ID.

        :param str reaction_id: The reaction identifier.
        :returns: A string representation of the GPR rule.

        """
        if reaction_id not in self.reactions:
            raise ValueError(f"Reactions {reaction_id} does not exist")
        reaction = self.model.reactions.get_by_id(reaction_id)
        if reaction.gene_reaction_rule:
            return str(reaction.gene_reaction_rule)
        else:
            return None

    def get_exchange_reactions(self):
        rxns = [r.id for r in self.model.exchanges]
        return rxns

    def get_gene_reactions(self):
        """
        :returns: A map of genes to reactions.
        """
        if not self._gene_to_reaction:
            gr = dict()
            for rxn_id in self.reactions:
                rxn = self.model.reactions.get_by_id(rxn_id)
                genes = rxn.genes
                for g in genes:
                    if g.id in gr.keys():
                        gr[g.id].append(rxn_id)
                    else:
                        gr[g.id] = [rxn_id]
            self._gene_to_reaction = gr
        return self._gene_to_reaction


class Simulation(CobraModelContainer, Simulator):
    """Generic Simulation class for cobra Model.
       Defines the simulation conditions, and makes available a set of methods.

    :param model: An metabolic model instance.

    Optional:

    :param dic envcond: Dictionary of environmental conditions.
    :param dic constraints: A dictionary of reaction constraints.
    :param solver: An instance of the LP solver.
    :param dic reference: A dictionary of the wild type flux values.

    """

    def __init__(self, model: Model, envcond=None, constraints=None, solver=None, reference=None,
                 reset_solver=ModelConstants.RESET_SOLVER):

        if not isinstance(model, Model):
            raise ValueError("Model is incompatible or inexistent")

        self.model = model
        self.model.solver = get_default_solver()
        self._environmental_conditions = OrderedDict() if envcond is None else envcond
        self._constraints = dict() if constraints is None else {
            k: v for k, v in constraints.items() if k not in list(self._environmental_conditions.keys())}

        self.solver = solver
        self._gene_to_reaction = None
        self._reference = reference
        self.__status_mapping = {
            'optimal': SStatus.OPTIMAL,
            'unbounded': SStatus.UNBOUNDED,
            'infeasible': SStatus.INFEASIBLE,
            'infeasible_or_unbounded': SStatus.INF_OR_UNB,
            'suboptimal': SStatus.SUBOPTIMAL,
            'unknown': SStatus.UNKNOWN
        }
        self.solver = solver
        self._reset_solver = reset_solver
        self.reverse_sintax = []
        self._m_r_lookup = None

        self._MAX_STR = 'maximize'
        self._MIN_STR = 'minimize'

        # apply the env. cond. and additional constraints to the model
        for r_id, bounds in self._environmental_conditions.items():
            self._set_model_reaction_bounds(r_id, bounds)
        for r_id, bounds in self._constraints.items():
            self._set_model_reaction_bounds(r_id, bounds)

    @property
    def environmental_conditions(self):
        return self._environmental_conditions.copy()

    @environmental_conditions.setter
    def environmental_conditions(self, a):
        raise ValueError("Can not change environmental conditions. Use set_reaction_bounds instead")

    def _set_model_reaction_bounds(self, r_id, bounds):
        if isinstance(bounds, tuple):
            lb = bounds[0]
            ub = bounds[1]
        elif isinstance(bounds, (int, float)):
            lb = bounds
            ub = bounds
        else:
            raise ValueError(f"Invalid bounds definition {bounds}")
        rxn = self.model.reactions.get_by_id(r_id)
        rxn.bounds = (lb, ub)

    @property
    def objective(self):
        from cobra.util.solver import linear_reaction_coefficients
        d = dict(linear_reaction_coefficients(self.model))
        return {k.id: v for k, v in d.items()}

    @objective.setter
    def objective(self, objective):
        if isinstance(objective, str):
            self.model.objective = objective
        elif isinstance(objective, dict):
            from cobra.util.solver import set_objective
            linear_coef = {self.model.reactions.get_by_id(r_id): v for r_id, v in objective.items()}
            set_objective(self.model, linear_coef)
        else:
            raise ValueError(
                'The objective must be a reaction identifier or a dictionary of \
                reaction identifier with respective coeficients.')

    def add_metabolite(self, id, formula=None, name=None, compartment=None):
        from cobra import Metabolite
        meta = Metabolite(id, formula=formula, name=name, compartment=compartment)
        self.model.add_metabolites([meta])

    def add_reaction(self, rxn_id, stoichiometry, lb=ModelConstants.REACTION_LOWER_BOUND,
                     ub=ModelConstants.REACTION_UPPER_BOUND, replace=True, *kwargs):
        """Adds a reaction to the model

        Args:
            rxn_id: The reaction identifier
            stoichiometry: The reaction stoichiometry, a dictionary of species: coefficient
            lb: Reaction flux lower bound, defaults to ModelConstants.REACTION_LOWER_BOUND
            ub: Reaction flux upper bound, defaults to ModelConstants.REACTION_UPPER_BOUND
            replace(bool, optional): If the reaction should be replaced in case it is already defined.\
                Defaults to True.
        """
        from cobra import Reaction
        reaction = Reaction(rxn_id)
        reaction.add_metabolites(stoichiometry)
        reaction.lower_bound = lb
        reaction.upper_bound = ub
        self.model.add_reaction(reaction)

    def remove_reaction(self, r_id):
        """Removes a reaction from the model.

        Args:
            r_id (str): The reaction identifier.
        """
        self.model.remove_reactions(r_id)

    def get_uptake_reactions(self):
        """
        :returns: The list of uptake reactions.
        """
        drains = self.get_exchange_reactions()
        rxns = [r for r in drains if self.model.reactions.get_by_id(r).reversibility
                or ((self.model.reactions.get_by_id(r).lower_bound is None
                     or self.model.reactions.get_by_id(r).lower_bound < 0)
                    and len(self.model.reactions.get_by_id(r).reactants) > 0)
                or ((self.model.reactions.get_by_id(r).upper_bound is None
                     or self.model.reactions.get_by_id(r).upper_bound > 0)
                    and len(self.model.reactions.get_by_id(r).products) > 0)
                ]
        return rxns

    def get_transport_reactions(self):
        """
        :returns: The list of transport reactions.
        """
        transport_reactions = []
        for rx in self.reactions:
            s_set = set()
            p_set = set()
            s = self.model.reactions.get_by_id(rx).reactants
            for x in s:
                s_set.add(x.compartment)
            p = self.model.reactions.get_by_id(rx).products
            for x in p:
                p_set.add(x.compartment)
            if len(p_set.intersection(s_set)) == 0:
                transport_reactions.append(rx)
        return transport_reactions

    def get_transport_genes(self):
        """Returns the list of genes that only catalyze transport reactions.
        """
        trp_rxs = self.get_transport_reactions()
        r_g = self.get_gene_reactions()
        genes = []
        for g, rxs in r_g.items():
            if set(rxs).issubset(set(trp_rxs)):
                genes.append(g)
        return genes

    def reverse_reaction(self, reaction_id):
        """
        Identify if a reaction is reversible and returns the reverse reaction if it is the case.

        :param reaction_id: A reaction identifier.
        :returns: A reaction identifier or None.

        """
        rxn = self.model.reactions.get_by_id(reaction_id)
        reactions = self.reactions
        if rxn.lower_bound < 0:
            return reaction_id
        else:
            rev = rxn.reverse_id
            if rev in reactions:
                return rev
        return None

    def metabolite_reaction_lookup(self, force_recalculate=False):
        """ Return the network topology as a nested map from metabolite to reaction to coefficient.
        :return: a dictionary lookup table
        """

        if not self._m_r_lookup or force_recalculate:
            self._m_r_lookup = OrderedDict([(m_id, OrderedDict()) for m_id in self.metabolites])
            for _, reaction in enumerate(self.model.reactions):
                for m, coeff in reaction.metabolites.items():
                    self._m_r_lookup[m.id][reaction.id] = coeff

        return self._m_r_lookup

    def metabolite_elements(self, metabolite_id):
        return self.model.metabolites.get_by_id(metabolite_id).elements

    def get_reaction_bounds(self, reaction_id):
        """
        Returns the bounds for a given reaction.

        :param reaction_id: str, reaction ID
        :return: lb(s), ub(s), tuple

        """
        lb, ub = self.model.reactions.get_by_id(reaction_id).bounds
        return lb if lb > -np.inf else ModelConstants.REACTION_LOWER_BOUND,\
            ub if ub < np.inf else ModelConstants.REACTION_UPPER_BOUND

    def set_reaction_bounds(self, reaction_id, lb=None, ub=None):
        """
        Sets the bounds for a given reaction.
        :param reaction_id: str, reaction ID
        :param float lb: lower bound 
        :param float ub: upper bound
        """
        if reaction_id in self.get_uptake_reactions():
            self._environmental_conditions[reaction_id] = (lb, ub)
        else:
            self._constraints[reaction_id] = (lb, ub)
        rxn = self.model.reactions.get_by_id(reaction_id)
        rxn.bounds = (lb, ub)

    def find_bounds(self):
        """
        Return the median upper and lower bound of the metabolic model.
        Bounds can vary from model to model. Cobrapy defaults to (-1000, 1000).
        """
        lower_bounds = np.asarray([rxn.lower_bound for rxn in self.model.reactions], dtype=float)
        upper_bounds = np.asarray([rxn.upper_bound for rxn in self.model.reactions], dtype=float)
        lower_bound = np.nanmedian(lower_bounds[lower_bounds != 0.0])
        upper_bound = np.nanmedian(upper_bounds[upper_bounds != 0.0])
        if np.isnan(lower_bound):
            LOGGER.warning("Could not identify a median lower bound.")
            lower_bound = ModelConstants.REACTION_LOWER_BOUND
        if np.isnan(upper_bound):
            LOGGER.warning("Could not identify a median upper bound.")
            upper_bound = ModelConstants.REACTION_UPPER_BOUND
        return lower_bound, upper_bound

    def find_unconstrained_reactions(self):
        """Return list of reactions that are not constrained at all."""
        lower_bound, upper_bound = self.find_bounds()
        return [
            rxn.id
            for rxn in self.model.reactions
            if rxn.lower_bound <= lower_bound and rxn.upper_bound >= upper_bound
        ]

    # The simulator
    def simulate(self, objective=None, method=SimulationMethod.FBA, maximize=True,
                 constraints=None, reference=None, scalefactor=None, solver=None, slim=False):
        '''
        Simulates a phenotype when applying a set constraints using the specified method.

        :param dic objective: The simulation objective. If none, the model objective is considered.
        :param method: The SimulationMethod (FBA, pFBA, lMOMA, etc ...)
        :param boolean maximize: The optimization direction.
        :param dic constraints: A dictionary of constraints to be applied to the model.
        :param dic reference: A dictionary of reaction flux values.
        :param float scalefactor: A positive scaling factor for the solver. Default None.

        '''

        if not objective:
            objective = self.model.objective
        elif isinstance(objective, dict) and len(objective) > 0:
            objective = next(iter(objective.keys()))

        simul_constraints = {}
        if constraints:
            simul_constraints.update({k: v for k, v in constraints.items()
                                      if k not in list(self._environmental_conditions.keys())})

        with self.model as model:
            model.objective = objective
            for rxn in list(simul_constraints.keys()):
                reac = model.reactions.get_by_id(rxn)
                # constraints defined as a tuple (lower_bound, upper_bound) or as a single float value
                if isinstance(simul_constraints.get(rxn), tuple):
                    reac.bounds = (simul_constraints.get(
                        rxn)[0], simul_constraints.get(rxn)[1])
                else:
                    reac.bounds = (simul_constraints.get(
                        rxn), simul_constraints.get(rxn))
            # NOTE: If working directly over optlang use 'max' and 'min'
            # such is the case with pytfa.core.Model... need to find some workaround
            objective_sense = self._MAX_STR if maximize else self._MIN_STR
            if method == SimulationMethod.FBA:
                if slim:
                    solution = model.slim_optimize()
                else:
                    solution = model.optimize(objective_sense=objective_sense)
            elif method == SimulationMethod.pFBA:
                # make fraction_of_optimum a configurable parameter?
                solution = pfba(model)
            elif method == SimulationMethod.lMOMA or method == SimulationMethod.MOMA:
                s = None
                if not maximize:
                    s = model.optimize(objective_sense=objective_sense)
                linear = True if method == SimulationMethod.lMOMA else False
                solution = moma(model, solution=s, linear=linear)
            elif method == SimulationMethod.ROOM:
                solution = room(model)
            # Special case in which only the simulation context is required without any simulation result
            elif method == SimulationMethod.NONE:
                solution = Solution(None, 'unknown', None)
                pass
            else:
                raise Exception(
                    "Unknown method to perform the simulation.")

        if slim:
            return solution

        else:
            status = self.__status_mapping[solution.status]
            result = SimulationResult(model, solution.objective_value, fluxes=solution.fluxes.to_dict(OrderedDict),
                                      status=status, envcond=self.environmental_conditions,
                                      model_constraints=self._constraints.copy(),
                                      simul_constraints=constraints,
                                      maximize=maximize,
                                      method=method)
            return result

    def FVA(self, obj_frac=0.9, reactions=None, constraints=None, loopless=False, internal=None, solver=None,
            format='dict'):
        """ Flux Variability Analysis (FVA).

        :param model: An instance of a constraint-based model.
        :param float obj_frac: The minimum fraction of the maximum growth rate (default 0.9). Requires that the \
            objective value is at least the fraction times maximum objective value. A value of 0.85 for instance \
            means that the objective has to be at least at 85% percent of its maximum.
        :param list reactions: List of reactions to analyze (default: all).
        :param dic constraints: Additional constraints (optional).
        :param boolean loopless: Run looplessFBA internally (very slow) (default: false).
        :param list internal: List of internal reactions for looplessFBA (optional).
        :param solver: A pre-instantiated solver instance (optional).
        :param format: The return format: 'dict', returns a dictionary,'df' returns a data frame.
        :returns: A dictionary of flux variation ranges.

        """
        from cobra.flux_analysis.variability import flux_variability_analysis

        simul_constraints = {}
        if constraints:
            simul_constraints.update({k: v for k, v in constraints.items()
                                      if k not in list(self._environmental_conditions.keys())})

        with self.model as model:

            if simul_constraints:
                for rxn in list(simul_constraints.keys()):
                    reac = model.reactions.get_by_id(rxn)
                    if isinstance(simul_constraints.get(rxn), tuple):
                        reac.bounds = (simul_constraints.get(
                            rxn)[0], simul_constraints.get(rxn)[1])
                    else:
                        reac.bounds = (simul_constraints.get(
                            rxn), simul_constraints.get(rxn))

            df = flux_variability_analysis(
                model, reaction_list=reactions, loopless=loopless, fraction_of_optimum=obj_frac)

        variability = {}
        for r_id in reactions:
            variability[r_id] = [
                float(df.loc[r_id][0]), float(df.loc[r_id][1])]

        if format == 'df':
            import pandas as pd
            e = variability.items()
            f = [[a, b, c] for a, [b, c] in e]
            df = pd.DataFrame(f, columns=['Reaction ID', 'Minimum', 'Maximum'])
            return df
        else:
            return variability

    def set_objective(self, reaction):
        self.model.objective = reaction


class GeckoSimulation(Simulation):
    """
    Simulator for geckopy.gecko.GeckoModel
    """

    def __init__(self, model, envcond=None, constraints=None, solver=None, reference=None,
                 reset_solver=ModelConstants.RESET_SOLVER, protein_prefix=None):
        try:
            from geckopy.gecko import GeckoModel
            if not isinstance(model, GeckoModel):
                raise ValueError("The model is not an instance of geckopy.gecko.GeckoModel")
        except ImportError:
            raise RuntimeError("The geckopy package is not installed.")

        super(GeckoSimulation, self).__init__(
            model, envcond, constraints, solver, reference, reset_solver)
        self.protein_prefix = protein_prefix if protein_prefix else 'draw_prot_'
        self._essential_proteins = None
        self._protein_rev_reactions = None

    @property
    def proteins(self):
        return list(self.model.proteins)

    def essential_proteins(self, min_growth=0.01):
        if self._essential_proteins is not None:
            return self._essential_proteins
        wt_solution = self.simulate()
        wt_growth = wt_solution.objective_value
        self._essential_proteins = []
        proteins = self.model.proteins
        for p in tqdm(proteins):
            rxn = "{}{}".format(self.protein_prefix, p)
            res = self.simulate(constraints={rxn: 0})
            if res:
                if (res.status == SStatus.OPTIMAL and res.objective_value < wt_growth * min_growth) or \
                        res.status == SStatus.INFEASIBLE:
                    self._essential_proteins.append(rxn)
        return self._essential_proteins

    def protein_reactions(self, protein):
        """
        Returns the list of reactions associated to a protein.
        """
        reactions = []
        for r_id, rxn in self.model.reactions.items():
            lsub = rxn.reactants
            for m in lsub:
                if protein in m:
                    reactions.append(r_id)
        return reactions

    def reverse_reaction(self, reaction_id):
        """
        Identify if a reaction is reversible and returns the reverse reaction if it is the case.

        :param reaction_id: A reaction identifier.
        :returns: The reverse reaction identifier if exists or None.
        """
        f, d = zip(*self.protein_rev_reactions.values())
        if reaction_id in f:
            return d[f.index(reaction_id)]
        elif reaction_id in d:
            return f[d.index(reaction_id)]
        else:
            return None

    @property
    def protein_rev_reactions(self):
        """
        Pairs of reverse reactions associated with a protein

        :returns: A dictionary which identifies for each protein (key) the list of reversible reactions pairs.

        """
        if not self._protein_rev_reactions:
            proteins = self.model.proteins
            reactions = self.reactions
            in_sub = {}
            for p in proteins:
                sub = []
                for r_id in reactions:
                    rxn = self.model.reactions.get_by_id(r_id)
                    lsub = rxn.reactants
                    for m in lsub:
                        if p in m.id:
                            sub.append(r_id)
                in_sub[p] = sub
            pairs = {}
            for k, s in in_sub.items():
                revs = [r for r in s if '_REV' in r]
                if len(revs) > 0:
                    for r in revs:
                        la = [a for a in s if r.replace('_REV', '') == a]
                        la.append(r)
                        if len(la) == 2:
                            if k in pairs.keys():
                                pairs[k].append((la[0], la[1]))
                            else:
                                pairs[k] = [(la[0], la[1])]
            self._protein_rev_reactions = pairs
        return self._protein_rev_reactions

    def getKcat(self, protein):
        """ Returns a dictionary of reactions and respective Kcat for a specific enzyme·
        """
        m_r = self.metabolite_reaction_lookup()
        r_d = m_r[protein]
        return {k: v for k, v in r_d.items() if v < 0}



from collections import OrderedDict
from libsbml import AssignmentRule, SBMLReader
import os
from ..model.kinetic import ODEModel, KineticReaction, Rule


def load_sbml(filename):
    """ Loads an SBML file.

    :param filename: SBML file path, str.
    :returns: SBMLModel

    """

    if not os.path.exists(filename):
        raise IOError("Model file was not found")

    reader = SBMLReader()
    document = reader.readSBML(str(filename))
    sbml_model = document.getModel()

    if sbml_model is None:
        document.printErrors()
        raise IOError(f'Failed to load model {filename}.')

    return sbml_model


def load_ODEModel(filename):
    sbml_model = load_sbml(filename)
    from reframed.io.sbml import load_compartments, load_metabolites, load_reactions
    ode_model = ODEModel(sbml_model.getId())
    load_compartments(sbml_model, ode_model)
    load_metabolites(sbml_model, ode_model)
    # adds constants and boundaries to metabolites
    for species in sbml_model.getListOfSpecies():
        m = ode_model.metabolites[species.getId()]
        if not hasattr(m, "constant"):
            m.constant = species.getConstant()
        if not hasattr(m, "boundary"):
            m.boundary = species.getBoundaryCondition()
    load_reactions(sbml_model, ode_model)
    _load_concentrations(sbml_model, ode_model)
    _load_global_parameters(sbml_model, ode_model)
    _load_ratelaws(sbml_model, ode_model)
    _load_assignment_rules(sbml_model, ode_model)
    return ode_model


def _load_concentrations(sbml_model, odemodel):
    for species in sbml_model.getListOfSpecies():
        odemodel.set_concentration(species.getId(), species.getInitialConcentration())


def _load_global_parameters(sbml_model, odemodel):
    for parameter in sbml_model.getListOfParameters():
        odemodel.set_global_parameter(parameter.getId(), parameter.getValue(), parameter.getConstant())


def _load_ratelaws(sbml_model, odemodel):
    for reaction in sbml_model.getListOfReactions():
        formula = reaction.getKineticLaw().getFormula()
        parameters = OrderedDict()
        substrates = []
        products = []
        modifiers = []
        for parameter in reaction.getKineticLaw().getListOfParameters():
            parameters[parameter.getId()] = parameter.getValue()
        for reactant in reaction.getListOfReactants():
            m_id = reactant.getSpecies()
            substrates.append(m_id)
        for product in reaction.getListOfProducts():
            m_id = product.getSpecies()
            products.append(m_id)
        for modifier in reaction.getListOfModifiers():
            m_id = modifier.getSpecies()
            modifiers.append(m_id)
        law = KineticReaction(reaction.getId(), formula, parameters=parameters,
                              substrates=substrates, products=products, modifiers=modifiers)
        odemodel.set_ratelaw(reaction.getId(), law)


def _load_assignment_rules(sbml_model, odemodel):
    for rule in sbml_model.getListOfRules():
        if isinstance(rule, AssignmentRule):
            r_id = rule.getVariable()
            odemodel.set_assignment_rule(r_id, Rule(r_id, rule.getFormula()))



from typing import Union
from io import TextIOWrapper

try:
    # noinspection PyPackageRequirements
    from cobra import Model as Cobra_Model

except ImportError:
    Cobra_Model = str

try:
    # noinspection PyPackageRequirements
    from reframed import CBModel as Reframed_Model

except ImportError:

    Reframed_Model = str

from ..io import Reader, Engines, read_model
from .problem import AbstractKOProblem


# TODO: should it be in io?
def load_optorf(regulatory_model: Union[str, TextIOWrapper, Reader],
                metabolic_model: Union[str, TextIOWrapper, Cobra_Model, Reframed_Model, Reader],
                config: dict = None,
                warnings: bool = False):
    if not config:
        config = {}

    if not isinstance(regulatory_model, Reader):

        if isinstance(regulatory_model, str):

            file_name = regulatory_model

        elif isinstance(regulatory_model, TextIOWrapper):

            file_name = regulatory_model.name

        else:
            raise ImportError('Invalid file type')

        engine = Engines.RegulatoryCSV

        if file_name.endswith('.xml') or file_name.endswith('.sbml'):
            engine = Engines.RegulatorySBML

        regulatory_model = Reader(engine=engine,
                                  io=regulatory_model,
                                  **config)

    if not isinstance(metabolic_model, Reader):

        if isinstance(metabolic_model, str):

            engine = Engines.MetabolicSBML

        elif isinstance(metabolic_model, TextIOWrapper):

            engine = Engines.MetabolicSBML

        elif isinstance(metabolic_model, Cobra_Model):

            engine = Engines.CobrapyModel

        elif isinstance(metabolic_model, Reframed_Model):

            engine = Engines.ReframedModel

        else:
            raise ImportError('Invalid file type')

        metabolic_model = Reader(engine=engine,
                                 io=metabolic_model,
                                 **config)

    return read_model(regulatory_model, metabolic_model, warnings=warnings)


class OptORFProblem(AbstractKOProblem):

    def __init__(self, model, fevaluation, **kwargs):

        """
        OptORF problem using the RFBA implementation analysis.
        The OptORF approach is based on gene and regulator deletion to identify optimization strategies.

        It currently supports KOProblem-based decoding

        For more details consult: https://doi.org/10.1186/1752-0509-4-53

        """

        # TODO: this is the main disadvantage of the new model integration.
        #  Integrated models (metabolic-regulatory) are only available with conversion to mewpy models.
        #  However, one can easily converted a reframed or cobra model to mewpy with the IO read that I implemented.
        if isinstance(model, (Cobra_Model, Reframed_Model)):

            raise ValueError(f'OptORF is not available for a model of type {type(model)}.'
                             f'Please use load_optorf() to retrieve an integrated mewpy model')

        super(OptORFProblem, self).__init__(model, fevaluation, **kwargs)

        # ignore user defined target list
        self._trg_list = None

        self._initial_state = {}

    def _build_target_list(self):

        # Target list is the combination of genes and regulators available into the mewpy integrated model

        regulators = [regulator.id for regulator in self.model.yield_regulators()
                      if not regulator.is_reaction() and not regulator.is_metabolite() and not regulator.is_gene()]

        self._trg_list = [gene for gene in self.model.genes] + regulators

        self._initial_state = {trg: 1 for trg in self._trg_list}

    def _sim_bool(self, state):

        # see RFBA implementation for details

        result = {}

        for interaction in self.model.yield_interactions():

            is_empty = True
            active_coefficient = None

            for coefficient, regulatory_event in interaction.regulatory_events.items():

                if not regulatory_event.is_none:

                    is_empty = False

                    regulatory_event_eval = regulatory_event.evaluate(values=state)

                    if regulatory_event_eval:
                        active_coefficient = coefficient

                        break

            if is_empty:
                target_val = interaction.target.coefficient.active_coefficient

            elif active_coefficient is not None:
                target_val = active_coefficient

            else:
                target_val = 0.0

            result[interaction.target.id] = target_val

        return result

    def _decode_metabolic_state(self, state):

        # see RFBA implementation for details

        constraints = {}

        # gpr over model.reactions
        for rxn in self.model.yield_reactions():

            if rxn.gpr.is_none:
                continue

            else:

                res = rxn.gpr.evaluate(values=state)

                if not res:
                    constraints[rxn.id] = (0.0, 0.0)

        return constraints

    def decode(self, candidate):

        """

        OptORF candidate decode

        :param candidate:
        :return:
        """

        for t_idx in candidate:

            self._initial_state[self.target_list[t_idx]] = 0

        # simulation of the regulatory network to determine the affected metabolic state
        state = self._sim_bool(self._initial_state)

        # find the affected reactions
        constraints = self._decode_metabolic_state(state)

        return constraints



from mewpy.solvers.solution import Status
from mewpy.util.constants import ModelConstants


def decode_solver_solution(solution, minimize=True, status=False):

    sol_status = solution.status.value

    if solution.status == Status.OPTIMAL:
        sol_f_obj = solution.fobj

    elif solution.status == Status.UNBOUNDED or solution.status == Status.INF_OR_UNB:

        if minimize:

            sol_f_obj = ModelConstants.REACTION_LOWER_BOUND

        else:

            sol_f_obj = ModelConstants.REACTION_UPPER_BOUND

    else:
        sol_f_obj = None

    if status:
        return sol_f_obj, sol_status

    return sol_f_obj



from tqdm import tqdm
import joblib
import contextlib
import functools
import re
import types
import time
from collections import Iterable


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""


class Timer:
    def __init__(self):
        self._start_time = None

    def start(self):
        """Start a new timer"""
        if self._start_time is not None:
            raise TimerError("Timer is running. Use .stop() to stop it")

        self._start_time = time.perf_counter()

    def stop(self):
        """Stop the timer, and report the elapsed time"""
        if self._start_time is None:
            raise TimerError("Timer is not running. Use .start() to start it")

        elapsed_time = time.perf_counter() - self._start_time
        self._start_time = None
        print(f"Elapsed time: {elapsed_time:0.6f} seconds")

    def __enter__(self):
        """Start a new timer as a context manager"""
        self.start()
        return self

    def __exit__(self, *exc_info):
        """Stop the context manager timer"""
        self.stop()


class Singleton(object):
    _instance = None

    def __new__(class_, *args, **kwargs):
        if not isinstance(class_._instance, class_):
            class_._instance = object.__new__(class_, *args, **kwargs)
        return class_._instance


def copy_func(f):
    """Based on http://stackoverflow.com/a/6528148/190597 (Glenn Maynard)"""
    g = types.FunctionType(f.__code__, f.__globals__, name=f.__name__,
                           argdefs=f.__defaults__,
                           closure=f.__closure__)
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    return g


class Dispatcher:

    def __init__(self):
        """
        Dispatcher for the simulate method of the Simulation interface
        It allows a simplification of the if else chain of methods provided as input to the simulate method

        based on https://stackoverflow.com/questions/36836161/singledispatch-based-on-value-instead-of-type
        (Ilja Everilä)

        """

        # weak ref key for the garbage collector
        self.registry = {}

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self.dispatch(instance, owner)

    def dispatch(self, instance, owner):
        def wrapper(state, *args, **kwargs):
            method = self.registry.get(state).__get__(instance, owner)
            return method(*args, **kwargs)
        return wrapper

    def register(self, state):
        def wrapper(method):
            self.registry[state] = method
            return method
        return wrapper


def iterable(obj, is_string=False):
    if isinstance(obj, Iterable):
        if is_string and isinstance(obj, str):
            return (obj,)
        return obj
    return (obj,)


def generator(container):
    return (value for value in container.values())


# Taken from the talented team responsible for developing cobrapy!!!!
chemical_formula_re = re.compile('([A-Z][a-z]?)([0-9.]+[0-9.]?|(?=[A-Z])?)')


def elements(formula):
    all_elements = re.findall(chemical_formula_re, formula)
    atoms = {}
    for atom, count in all_elements:
        if not count:
            count = '1'
        atoms[atom] = atoms.get(atom, 0) + int(count)
    return atoms


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""
    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()



import re
from io import StringIO
from token import NAME, OP, NUMBER
from tokenize import generate_tokens, untokenize

from .symbolic import NoneAtom, Symbolic
from .algebra_constants import (BOOLEAN_STATES,
                                BOOLEAN_OPERATORS,
                                TRUE,
                                FALSE,
                                RELATIONAL_STATES,
                                RELATIONAL_OPERATORS,
                                RELATIONAL_EQUAL_OPERATORS,
                                GLOBAL_RELATIONAL_EQUAL_OPERATORS,
                                GLOBAL_RELATIONAL_OPERATORS,
                                BOOLEAN_ESCAPE_CHARS,
                                RELATIONAL_ESCAPE_CHARS,
                                GLOBAL_MEWPY_OPERATORS)

_relational_ops = set()
_relational_ops.update(GLOBAL_RELATIONAL_OPERATORS)
_relational_ops.update(GLOBAL_RELATIONAL_EQUAL_OPERATORS)


class ExpressionParser:

    def __init__(self,
                 expression=None,

                 stringify_expression='',
                 parsed_expression='',
                 symbolic_expression='',
                 tokenized_expression=None,

                 tokens=None,
                 filters=None,
                 transformations=None,
                 escape_chars=None,
                 replaces=None,

                 symbols=None,
                 aliases=None,

                 is_boolean=False,
                 is_true=False,
                 is_false=False,
                 is_and=False,
                 is_or=False,
                 is_not=False,

                 is_relational=False,
                 is_equal=False,
                 is_not_equal=False,
                 is_inequality=False,
                 is_greater=False,
                 is_greater_equal=False,
                 is_less=False,
                 is_less_equal=False,

                 is_numeric=False,
                 is_integer=False,
                 is_float=False,
                 is_one=False,
                 is_zero=False,

                 is_symbol=False,
                 is_atom=False,

                 is_none=False):

        """
        Internal use
        """

        if aliases is None:
            aliases = {}

        if symbols is None:
            symbols = {}

        if replaces is None:
            replaces = {}

        if escape_chars is None:
            escape_chars = {}

        if tokens is None:
            tokens = []

        if filters is None:
            filters = []

        if transformations is None:
            transformations = []

        if tokenized_expression is None:
            tokenized_expression = []

        self.expression = expression

        self.stringify_expression = stringify_expression
        self.parsed_expression = parsed_expression
        self.symbolic_expression = symbolic_expression
        self.tokenized_expression = tokenized_expression

        self.tokens = tokens
        self.filters = filters
        self.transformations = transformations
        self.escape_chars = escape_chars
        self.replaces = replaces

        self.symbols = symbols
        self.aliases = aliases

        self.is_boolean = is_boolean
        self.is_true = is_true
        self.is_false = is_false
        self.is_and = is_and
        self.is_or = is_or
        self.is_not = is_not

        self.is_relational = is_relational
        self.is_equal = is_equal
        self.is_not_equal = is_not_equal
        self.is_inequality = is_inequality
        self.is_greater = is_greater
        self.is_greater_equal = is_greater_equal
        self.is_less = is_less
        self.is_less_equal = is_less_equal

        self.is_numeric = is_numeric
        self.is_integer = is_integer
        self.is_float = is_float
        self.is_one = is_one
        self.is_zero = is_zero

        self.is_symbol = is_symbol
        self.is_atom = is_atom

        self.is_none = is_none

    def build(self):

        for token in self.tokenized_expression:

            if token.lower() in BOOLEAN_OPERATORS:

                if self.is_boolean:
                    continue

                self.is_boolean = True
                diff = set(boolean_filters) - set(self.filters)
                self.filters = self.filters + list(diff)
                diff = set(boolean_transformations) - set(self.transformations)
                self.transformations = self.transformations + list(diff)
                self.escape_chars.update(BOOLEAN_ESCAPE_CHARS)

            elif token.lower() in RELATIONAL_OPERATORS or token.lower() in RELATIONAL_EQUAL_OPERATORS:

                if self.is_relational:
                    continue

                self.is_relational = True
                diff = set(relational_filters) - set(self.filters)
                self.filters = self.filters + list(diff)
                diff = set(relational_transformations) - set(self.transformations)
                self.transformations = self.transformations + list(diff)
                self.escape_chars.update(RELATIONAL_ESCAPE_CHARS)

            elif token.lower() in BOOLEAN_STATES or token.lower() in RELATIONAL_STATES:

                if self.is_one or self.is_true or self.is_false or self.is_zero:
                    continue

                bool_state = BOOLEAN_STATES.get(token.lower(), token.lower())

                if bool_state == TRUE:
                    self.is_one = True
                    self.is_true = True

                if bool_state == FALSE:
                    self.is_zero = True
                    self.is_false = True

                diff = set(state_filters) - set(self.filters)
                self.filters = self.filters + list(diff)
                diff = set(state_transformations) - set(self.transformations)
                self.transformations = self.transformations + list(diff)

            else:

                if self.is_symbol:
                    continue

                self.is_symbol = True
                diff = set(symbol_filters) - set(self.filters)
                self.filters = self.filters + list(diff)
                diff = set(symbol_transformations) - set(self.transformations)
                self.transformations = self.transformations + list(diff)
                self.escape_chars.update({**BOOLEAN_ESCAPE_CHARS, **RELATIONAL_ESCAPE_CHARS})

        if not self.is_boolean or self.is_relational or self.is_one or self.is_true or \
                self.is_zero or self.is_false or self.is_symbol:
            self.filters = all_filters
            self.transformations = all_transformations
            self.escape_chars = {**BOOLEAN_ESCAPE_CHARS, **RELATIONAL_ESCAPE_CHARS}


def tokenize(rule):
    return list(filter(lambda x: x != '', rule.replace('(', ' ( ').replace(')', ' ) ').split(' ')))


def escape_chars_filter(expression):
    """

    For a given self it parses out the special chars by replacing them with the corresponding values in the
    special_chars argument. By default, the global dictionary BOOL_SPECIAL_CHARS is used. This global dictionary
    can be altered for adding or removing special chars.

    """

    for escape_char, replace in expression.escape_chars.items():

        if escape_char in expression.parsed_expression:
            expression.replaces[replace] = escape_char
            expression.parsed_expression = expression.parsed_expression.replace(escape_char, replace)


def digit_filter(expression):
    """
    For a given regulatory self it checks if all regulatory variables start with a digit (str.is_digit()) and
    parses out the regulatory variable by adding the prefix '_dg_'

    """

    expression.parsed_expression = '(' + expression.parsed_expression + ')'
    regexp = re.compile(r'[^a-zA-Z|^_][0-9]+[a-zA-Z]|[^a-zA-Z|^_][0-9]+_')
    res = list(regexp.finditer(expression.parsed_expression))

    new_rule = ''

    last_nd = 0
    for i in range(len(res)):
        st = res[i].start() + 1
        nd = res[i].end()
        new_rule = new_rule + expression.parsed_expression[last_nd:st] + '_dg_' + expression.parsed_expression[
                                                                                  st:nd]
        last_nd = res[i].end()
        expression.replaces['_dg_'] = ''

    new_rule = new_rule + expression.parsed_expression[last_nd:]

    expression.parsed_expression = new_rule[1:-1]


def bitwise_filter(expression):
    """
    Transforms python boolean operators and, or and not, or other propositional logic syntax (e.g. greater
    than) to corresponding bitwise ones
    """

    bit_replace = {}

    if expression.is_boolean:
        bit_replace.update(BOOLEAN_OPERATORS)

    if expression.is_relational:
        bit_replace.update(RELATIONAL_EQUAL_OPERATORS)

    if expression.is_one or expression.is_true or expression.is_zero or expression.is_false:
        bit_replace.update(RELATIONAL_STATES)
        bit_replace.update(BOOLEAN_STATES)

    bit_ = {}

    for key, val in bit_replace.items():

        if key == val:
            continue

        if val in bit_:
            bit_[val].update({key, key.upper(), key.lower(), key.title()})
        else:
            bit_[val] = {key, key.upper(), key.lower(), key.title()}

    for bit_val, bools in bit_.items():

        for python_bool in bools:
            expression.parsed_expression = re.sub(r'\b{}\b'.format(python_bool), bit_val,
                                                  expression.parsed_expression)


def relational_filter(expression):
    """
    For a given boolean self (string) finds all relational expressions and encloses all of them with
    parenthesis and replaces the ambiguous operators

    """

    # regex to match a relational of any type: x>0; x<0; x>=0; x=>0;reverse order 0>x ...; with or without white
    # spaces ([a-zA-Z0-9_]+>[0-9]+)|([a-zA-Z0-9_]+\str>\str[0-9]+)

    regex_str = r'|'.join([r'([a-zA-Z0-9_]+' + regex + r'[0-9.]+)' + r'|' +
                           r'([a-zA-Z0-9_]+\s' + regex + r'\s[0-9.]+)' + r'|' +
                           r'([0-9.]+' + regex + r'[a-zA-Z0-9_]+)' + r'|' +
                           r'([0-9.]+\s' + regex + r'\s[a-zA-Z0-9_]+)'
                           for regex in _relational_ops])

    regexp = re.compile(regex_str)
    res = list(regexp.finditer(expression.parsed_expression))

    for match in set(map(lambda x: x.group(), res)):
        expression.parsed_expression = re.sub(r'\b{}\b'.format(match), '(' + match + ')',
                                              expression.parsed_expression)


def symbolic_transform(expression):
    """
    Heavily inspired by the talented people contributing for the impressive sympy package

    :param expression:
    :return:
    """

    result = []
    previous_token = (None, None)

    expression.tokens.append((None, None))  # so zip traverses all tokens

    for tok, next_token in zip(expression.tokens, expression.tokens[1:]):

        token_number, token_val = tok
        next_tok_num, next_tok_val = next_token

        if token_number == NAME:

            name = token_val

            if name == 'False':
                result.extend([(NAME, 'BoolFalse'),
                               (OP, '('),
                               (NAME, repr(str(name))),
                               (OP, ')')])

            elif name == 'True':
                result.extend([(NAME, 'BoolTrue'),
                               (OP, '('),
                               (NAME, repr(str(name))),
                               (OP, ')')])

            elif name == 'None':
                result.extend([(NAME, 'NoneAtom'),
                               (OP, '('),
                               (NAME, repr(str(name))),
                               (OP, ')')])

            else:
                result.extend([
                    (NAME, 'Symbol'),
                    (OP, '('),
                    (NAME, repr(str(token_val))),
                    (OP, ')'),
                ])

        else:

            result.append((token_number, token_val))

    expression.tokens = result


# FIXME: parsing does not handle minus so far. BOOLEAN_ESCAPE_CHARS = {'-': '_dash_',
# FIXME: parsing does not handle scientific notation so far. BOOLEAN_ESCAPE_CHARS = {'-': '_dash_',
def numeric_transform(expression):
    """

    Heavily inspired by the talented people contributing for the impressive sympy package

    :param expression:
    :return:
    """

    result = []

    for token_number, token_val in expression.tokens:

        if token_number == NUMBER:

            number = token_val

            if number in ('1.0', '1'):

                seq = [(NAME, 'One'),
                       (OP, '('),
                       (NUMBER, repr(str(number))),
                       (OP, ')')]

            elif number in ('0.0', '0'):

                seq = [(NAME, 'Zero'),
                       (OP, '('),
                       (NUMBER, repr(str(number))),
                       (OP, ')')]

            elif '.' in number or (('e' in number or 'E' in number)
                                   and not (number.startswith('0x') or number.startswith('0X'))):

                seq = [(NAME, 'Float'),
                       (OP, '('),
                       (NUMBER, repr(str(number))),
                       (OP, ')')]

            else:

                seq = [(NAME, 'Integer'),
                       (OP, '('),
                       (NUMBER, number),
                       (OP, ')')]

            result.extend(seq)

        else:

            result.append((token_number, token_val))

    expression.tokens = result


all_filters = [escape_chars_filter, digit_filter, bitwise_filter, relational_filter]
all_transformations = [symbolic_transform, numeric_transform]

boolean_filters = [escape_chars_filter, digit_filter, bitwise_filter]
boolean_transformations = [symbolic_transform, numeric_transform]

relational_filters = [escape_chars_filter, digit_filter, bitwise_filter, relational_filter]
relational_transformations = [symbolic_transform, numeric_transform]

symbol_filters = [escape_chars_filter, digit_filter]
symbol_transformations = [symbolic_transform]

state_filters = [bitwise_filter]
state_transformations = [symbolic_transform, numeric_transform]


def apply_filters(expression, filters=None):
    if not filters:
        filters = expression.filters

    sorted_filters = []

    for _filter in sorted_filters:
        if callable(_filter):
            if _filter is escape_chars_filter:
                sorted_filters.insert(0, _filter)
            else:
                sorted_filters.append(_filter)

    for _filter in filters:
        _filter(expression)

    return expression


def apply_transformations(expression, transformations=None):
    """
    Heavily inspired by the talented people contributing for the impressive sympy package

    :param expression:
    :param transformations:
    :return:
    """

    if not transformations:
        transformations = expression.transformations

    sorted_transformations = []

    for transform in transformations:
        if callable(transform):
            if transform is symbolic_transform:
                sorted_transformations.insert(0, transform)
            else:
                sorted_transformations.append(transform)

    for transform in sorted_transformations:
        transform(expression)

    expression.symbolic_expression = untokenize(expression.tokens)

    return expression


def evaluate_expression(symbolic_expression, local_dict=None, global_dict=None):
    if not local_dict:
        local_dict = {}

    if not global_dict:
        global_dict = GLOBAL_MEWPY_OPERATORS

    return eval(symbolic_expression, global_dict, local_dict)


# TODO: stubs and type hinting
def parse_expression(expression: str) -> Symbolic:
    if not expression:

        expr_parser = ExpressionParser()
        expr_parser.expression = NoneAtom()
        expr_parser.is_none = True

    elif isinstance(expression, str):

        expr_parser = ExpressionParser(expression=None,
                                       stringify_expression=expression,
                                       parsed_expression=expression,
                                       tokenized_expression=tokenize(expression))

        expr_parser.build()

        expr_parser = apply_filters(expr_parser)

        expr_parser.tokenized_expression = tokenize(expr_parser.parsed_expression)

        input_code = StringIO(expr_parser.parsed_expression.strip())

        for token_number, token_val, _, _, _ in generate_tokens(input_code.readline):
            expr_parser.tokens.append((token_number, token_val))

        expr_parser = apply_transformations(expr_parser)

        if len(expr_parser.tokenized_expression) < 1:
            expr_parser.expression = NoneAtom()
            expr_parser.is_none = True

        else:

            expr_parser.expression = evaluate_expression(expr_parser.symbolic_expression)

    else:

        raise ValueError('Expression could not be parsed')

    for element in expr_parser.expression:

        if element.is_symbol:

            old_reg_name = ''.join(element.value)

            for replace, special_char in expr_parser.replaces.items():
                old_reg_name = old_reg_name.replace(replace, special_char)

            element.name = old_reg_name

    return expr_parser.expression



from .expression import Expression
from .symbolic import *
from .parsing import parse_expression
from .algebra_utils import solution_decode

# TODO: I have implemented an algebra sub-package that allows smooth modeling operations with variables
#  and algebra expressions of any kind. It is similar to sympy expression handling. It is also similar to
#  the boolean syntax parser and probably slower. However, it allows me to handle variables in this expressions
#  and so dynamically retrieve regulators/genes objects for reactions/interactions/targets in my models.
#  I don't know if there is an easy way to combine all expression parsing. Thus, I should create a method to create a
#  Node of a boolean syntax parser from my symbolic mathematics



from typing import TYPE_CHECKING, Any, Union, Generator, Dict, List, Tuple, Set

from mewpy.model.model import Model
from mewpy.mew.lp import Notification
from mewpy.util.history import recorder
from mewpy.util.serialization import serialize
from mewpy.util.utilities import iterable, generator

if TYPE_CHECKING:
    from mewpy.mew.algebra import Expression
    from mewpy.mew.variables import Gene, Metabolite, Reaction


# TODO: methods stubs
class MetabolicModel(Model, model_type='metabolic', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 compartments: Dict[str, str] = None,
                 genes: Dict[str, 'Gene'] = None,
                 metabolites: Dict[str, 'Metabolite'] = None,
                 objective: Dict['Reaction', Union[float, int]] = None,
                 reactions: Dict[str, 'Reaction'] = None,
                 **kwargs):

        """
        A metabolic model contains essentially reactions, metabolites and genes.
        The metabolic model can be also also associated with a given objective for the simulation/analysis.
        The metabolic model can be loaded with compartments, although these can be inferred from the available
        metabolites.

        Other information is retrieved from these attributes, namely demand, exchange and sink reactions as well as gprs

        The metabolic model, as with other models, provides a clean interface for manipulation with the add, remove and
        update methods.

        :param identifier: identifier, e.g. iMC1010
        :param compartments: a dictionary with additional compartments not encoded in the metabolites
        :param genes: a dictionary with Gene objects. See variables.Gene for more info
        :param metabolites: a dictionary with Metabolite objects. See variables.Metabolite for more info
        :param objective: a dictionary with the Reaction objects that must be considered objective functions of
        the simulations together with the respective coefficients
        :param reactions: a dictionary with Reaction objects. See variables.Reaction for more info
        """

        # compartments attribute can be shared across the children, thus name mangling
        self.__compartments = {}
        self._genes = {}
        self._metabolites = {}
        self._objective = {}
        self._reactions = {}

        super().__init__(identifier,
                         **kwargs)

        # the setters will handle adding and removing variables to the correct containers
        self.compartments = compartments
        self.genes = genes
        self.metabolites = metabolites
        self.objective = objective
        self.reactions = reactions

    # -----------------------------------------------------------------------------
    # Model type manager
    # -----------------------------------------------------------------------------

    @serialize('types', None)
    @property
    def types(self):

        # noinspection PyUnresolvedReferences
        _types = {MetabolicModel.model_type}

        _types.update(super(MetabolicModel, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('genes', 'genes', '_genes')
    @property
    def genes(self) -> Dict[str, 'Gene']:
        return self._genes.copy()

    @serialize('metabolites', 'metabolites', '_metabolites')
    @property
    def metabolites(self) -> Dict[str, 'Metabolite']:
        return self._metabolites.copy()

    @serialize('objective', 'objective', '_objective')
    @property
    def objective(self) -> Dict['Reaction', Union[float, int]]:
        return self._objective.copy()

    @serialize('reactions', 'reactions', '_reactions')
    @property
    def reactions(self) -> Dict[str, 'Reaction']:
        return self._reactions.copy()

    @property
    def compartments(self) -> Dict[str, str]:

        compartments = {met.compartment: self.__compartments.get(met.compartment, '')
                        for met in self.yield_metabolites()
                        if met.compartment is not None}

        compartments.update(self.__compartments)

        compartments.update(super(MetabolicModel, self).compartments)

        return compartments

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------

    @compartments.setter
    @recorder
    def compartments(self, value: Dict[str, str]):

        if not value:
            value = {}

        self.__compartments.update(value)

    @genes.setter
    @recorder
    def genes(self, value: Dict[str, 'Gene']):

        if not value:
            value = {}

        self.remove(list(self.yield_genes()), 'gene', history=False)
        self.add(list(value.values()), 'gene', history=False)

    @metabolites.setter
    @recorder
    def metabolites(self, value: Dict[str, 'Metabolite']):

        if not value:
            value = {}

        self.remove(list(self.yield_metabolites()), 'metabolite', history=False)
        self.add(list(value.values()), 'metabolite', history=False)

    @objective.setter
    @recorder
    def objective(self, value: Dict['Reaction', Union[float, int]]):

        if not value:
            value = {}

        if isinstance(value, str):

            value = {self.get(value): 1}

        elif hasattr(value, 'types'):

            value = {value: 1}

        elif isinstance(value, dict):

            value = {self.get(var, var): val for var, val in value.items()}

        else:
            raise ValueError(f'{value} is not a valid objective')

        self._objective = value

        linear_obj = {var.id: coef for var, coef in self._objective.items()}

        content = {'linear': linear_obj,
                   'minimize': False}

        notification = Notification(content=content,
                                    content_type='objectives',
                                    action='set')

        self.notify(notification)

    @reactions.setter
    @recorder
    def reactions(self, value: Dict[str, 'Reaction']):

        if not value:
            value = {}

        self.remove(list(self.yield_reactions()), 'reaction', history=False)
        self.add(list(value.values()), 'reaction', history=False)

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------

    @property
    def external_compartment(self) -> Union[str, None]:

        if not self.compartments:
            return

        if not self._reactions:
            return

        boundary_compartments = {}

        for rxn in self.yield_reactions():

            if rxn.boundary:

                for compartment in rxn.compartments:
                    boundary_compartments[compartment] = boundary_compartments.get(compartment, 0) + 1

        external_compartment = None
        highest_count = 0

        for compartment, count in boundary_compartments.items():

            if count > highest_count:
                external_compartment = compartment
                highest_count = count

        return external_compartment

    def _get_boundaries(self):

        external_compartment = self.external_compartment

        if external_compartment is None:
            return {}, {}, {}

        all_boundaries = [rxn for rxn_id, rxn in self._reactions.items()
                          if rxn.boundary]

        exchanges = {}
        sinks = {}
        demands = {}

        for variable in all_boundaries:

            if variable.types == {'reaction'}:

                if external_compartment in variable.compartments:
                    exchanges[variable.id] = variable

                else:
                    if variable.reversibility:
                        sinks[variable.id] = variable

                    else:
                        demands[variable.id] = variable

        return exchanges, sinks, demands

    @property
    def demands(self) -> Dict[str, 'Reaction']:

        _, _, demands = self._get_boundaries()

        return demands

    @property
    def exchanges(self) -> Dict[str, 'Reaction']:

        exchanges, _, _ = self._get_boundaries()

        return exchanges

    @property
    def sinks(self) -> Dict[str, 'Reaction']:

        _, sinks, _ = self._get_boundaries()

        return sinks

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------

    def yield_compartments(self) -> Generator[str, None, None]:

        return generator(self.compartments)

    def yield_demands(self) -> Generator['Reaction', None, None]:

        return generator(self.demands)

    def yield_exchanges(self) -> Generator['Reaction', None, None]:

        return generator(self.exchanges)

    def yield_genes(self) -> Generator['Gene', None, None]:

        return generator(self._genes)

    def yield_gprs(self) -> Generator['Expression', None, None]:

        return (value.gpr for value in self._reactions.values())

    def yield_metabolites(self) -> Generator['Metabolite', None, None]:

        return generator(self._metabolites)

    def yield_reactions(self) -> Generator['Reaction', None, None]:

        return generator(self._reactions)

    def yield_sinks(self) -> Generator['Reaction', None, None]:

        return generator(self.sinks)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    def get(self, identifier: Any, default=None) -> Union['Gene', 'Metabolite', 'Reaction']:

        if identifier in self._metabolites:
            return self._metabolites[identifier]

        elif identifier in self._reactions:
            return self._reactions[identifier]

        elif identifier in self._genes:
            return self._genes[identifier]

        else:
            return super(MetabolicModel, self).get(identifier=identifier, default=default)

    def add(self,
            variables: Union[List[Union['Gene', 'Metabolite', 'Reaction']],
                             Tuple[Union['Gene', 'Metabolite', 'Reaction']],
                             Set[Union['Gene', 'Metabolite', 'Reaction']]],
            *types: str,
            comprehensive: bool = True,
            history=True):

        variables = iterable(variables)

        if not types:
            types = [var.types for var in variables]

        elif len(types) == len(variables):
            types = [{_type} if isinstance(_type, str) else set(_type) for _type in types]

        elif len(types) != len(variables):
            types = [set(types) for var in variables]

        reactions = []
        new_variables = []
        new_types = []

        for var_types, var in zip(types, variables):

            if 'gene' in var_types:
                self._add_gene(var)
                var_types.remove('gene')

            elif 'metabolite' in var_types:
                self._add_metabolite(var)
                var_types.remove('metabolite')

            elif 'reaction' in var_types:
                self._add_reaction(var, comprehensive=comprehensive)
                reactions.append(var)
                var_types.remove('reaction')

            if var_types:
                new_types.append(var_types)
                new_variables.append(var)

        if reactions:

            notification = Notification(content=reactions,
                                        content_type='reactions',
                                        action='add')

            self.notify(notification)

            notification = Notification(content=reactions,
                                        content_type='gprs',
                                        action='add')

            self.notify(notification)

        if history:
            self.history.queue_command(undo_func=self.remove,
                                       undo_kwargs={'variables': variables,
                                                    'remove_orphans': True,
                                                    'history': False},
                                       func=self.add,
                                       kwargs={'variables': variables,
                                               'comprehensive': comprehensive,
                                               'history': history})

        super(MetabolicModel, self).add(new_variables,
                                        *new_types,
                                        comprehensive=comprehensive,
                                        history=False)

    def remove(self,
               variables: Union[List[Union['Gene', 'Metabolite', 'Reaction']],
                                Tuple[Union['Gene', 'Metabolite', 'Reaction']],
                                Set[Union['Gene', 'Metabolite', 'Reaction']]],
               *types: str,
               remove_orphans: bool = False, history=True):

        variables = iterable(variables)

        if not types:
            types = [var.types for var in variables]

        elif len(types) == len(variables):
            types = [{_type} if isinstance(_type, str) else set(_type) for _type in types]

        elif len(types) != len(variables):
            types = [set(types) for var in variables]

        reactions = []
        metabolites = []
        new_variables = []
        new_types = []

        for var_types, var in zip(types, variables):

            if 'gene' in var_types:
                self._remove_gene(var)
                var_types.remove('gene')

            elif 'metabolite' in var_types:
                metabolites.append(var)
                self._remove_metabolite(var)
                var_types.remove('metabolite')

            elif 'reaction' in var_types:
                reactions.append(var)
                self._remove_reaction(var)
                var_types.remove('reaction')

            if var_types:
                new_types.append(var_types)
                new_variables.append(var)

        if reactions:

            notification = Notification(content=reactions,
                                        content_type='reactions',
                                        action='remove')

            self.notify(notification)

            notification = Notification(content=reactions,
                                        content_type='gprs',
                                        action='remove')

            self.notify(notification)

            if remove_orphans:
                orphan_mets, _ = self._remove_metabolic_orphans(reactions)

                notification = Notification(content=orphan_mets,
                                            content_type='metabolites',
                                            action='remove')

                self.notify(notification)

        # metabolites take precedence since linear coefficients are added to linear problems
        # by the reactions stoichiometry dictionary.
        # So, if metabolites are being removed as a result of removing reactions and its metabolites
        # all together from the model, the linear coefficients obtained from the reactions should be the last thing to
        # persist in the lp.
        if metabolites:
            notification = Notification(content=metabolites,
                                        content_type='metabolites',
                                        action='remove')

            self.notify(notification)

        if history:
            self.history.queue_command(undo_func=self.add,
                                       undo_kwargs={'variables': variables,
                                                    'comprehensive': True,
                                                    'history': False},
                                       func=self.remove,
                                       kwargs={'variables': variables,
                                               'remove_orphans': remove_orphans,
                                               'history': history})

        super(MetabolicModel, self).remove(new_variables,
                                           *new_types,
                                           remove_orphans=remove_orphans,
                                           history=False)

    def update(self,
               compartments: Dict[str, str] = None,
               objective: Dict['Reaction', Union[float, int]] = None,
               variables: Union[List[Union['Gene', 'Metabolite', 'Reaction']],
                                Tuple[Union['Gene', 'Metabolite', 'Reaction']],
                                Set[Union['Gene', 'Metabolite', 'Reaction']]] = None,
               **kwargs):

        if compartments is not None:
            self.compartments = compartments

        if variables is not None:
            self.add(variables=variables)

        if objective is not None:
            self.objective = objective

        super(MetabolicModel, self).update(**kwargs)

    # -----------------------------------------------------------------------------
    # Helper functions for the operations/Manipulations
    # -----------------------------------------------------------------------------

    def _add_reaction(self, reaction, comprehensive=True):

        # adding a reaction is regularly a in-depth append method, as both metabolites and genes associated with the
        # reaction are also regularly added to the model. Although, this behaviour can be avoided passing
        # comprehensive=False

        if reaction.id not in self._reactions:

            if comprehensive:

                for metabolite in reaction.yield_metabolites():
                    self._add_metabolite(metabolite)

                for gene in reaction.yield_genes():
                    self._add_gene(gene)

            self._add_variable_to_container(reaction, self._reactions)

    def _add_metabolite(self, metabolite):

        if metabolite.id not in self._metabolites:
            self._add_variable_to_container(metabolite, self._metabolites)

    def _add_gene(self, gene):

        if gene.id not in self._genes:
            self._add_variable_to_container(gene, self._genes)

    def _remove_reaction(self, reaction, remove_orphans=False):

        if reaction.id in self._reactions:
            self._remove_variable_from_container(reaction, self._reactions)

    def _remove_metabolite(self, metabolite):

        if metabolite.id in self._metabolites:
            self._remove_variable_from_container(metabolite, self._metabolites)

    def _remove_gene(self, gene):

        if gene.id in self._genes:
            self._remove_variable_from_container(gene, self._genes)

    def _remove_metabolic_orphans(self, reactions):

        orphan_mets = self._get_orphans(to_remove=reactions,
                                        first_container='metabolites',
                                        second_container='reactions')

        orphan_genes = self._get_orphans(to_remove=reactions,
                                         first_container='genes',
                                         second_container='reactions')

        if orphan_mets:

            for met in orphan_mets:
                self._remove_metabolite(met)

        if orphan_genes:
            for gene in orphan_genes:
                self._remove_gene(gene)

        return orphan_mets, orphan_genes



from typing import Union
from warnings import warn

from mewpy.solvers.solver import Solver
from mewpy.model import Model, MetabolicModel, RegulatoryModel
from mewpy.util.constants import ModelConstants
from mewpy.mew.lp import MetabolicLinearizer
from mewpy.mew.solution import ModelSolution, DynamicSolution


# TODO: type hinting and documentation
class RFBA(MetabolicLinearizer):

    def __init__(self,
                 model: Union[Model, MetabolicModel, RegulatoryModel],
                 solver: Union[str, Solver, None] = None,
                 build: bool = True,
                 attach: bool = False):
        """
        Regulatory Flux Balance Analysis (RFBA) of a metabolic-regulatory model.
        Implementation of a steady-state and dynamic versions of RFBA for a integrated metabolic-regulatory model.

        For more details consult Covert et al. 2004 at https://doi.org/10.1038/nature02456

        :param model: a mewpy Model, MetabolicModel, RegulatoryModel or all. The model is used to retrieve
        variables and constraints to the linear problem

        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used but it will be overwritten
        if build is true.

        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False
        """

        self._regulatory_reactions = []
        self._regulatory_metabolites = []

        super().__init__(model=model,
                         solver=solver,
                         build=build,
                         attach=attach)

    def build(self):

        if self._variables or self._constraints:
            self.clean()

        if self.model.is_metabolic():
            self.metabolite_reaction_lookup(reactions=self.model.yield_reactions(),
                                            metabolites=self.model.yield_metabolites())

            self._regulatory_reactions = [rxn.id
                                          for rxn in self.model.yield_reactions()
                                          if rxn.is_regulator()]

            self._regulatory_metabolites = [met.id
                                            for met in self.model.yield_metabolites()
                                            if met.is_regulator()]

            self.update()

            self.set_objective(linear={var.id: value for var, value in self.model.objective.items()},
                               minimize=False)

    def initial_state(self):

        if self.model.is_regulatory():
            return {regulator.id: regulator.coefficient.active_coefficient
                    for regulator in self.model.yield_regulators()}

        return {}

    def sim_bool(self, state):
        """
        Solves the boolean regulatory network for a given specific state.
        It also updates all targets having a valid regulatory interaction associated with it for the resulting state

        :param state: dict of regulatory variable keys (regulators) and a value of 0, 1 or float
        (reactions and metabolites predicates)

        :return: dict of target keys and a value of the resulting state
        """

        result = {}

        # solving regulatory rule by regulatory rule (synchronously though, as asynchronously would take too much time)
        # Each target has one and only one regulatory interaction
        # The regulatory model object only supports this strategy, though this can easily be altered. If so,
        # this solve method must be changed too

        if self.model.is_regulatory():

            # interaction over model.interactions.
            for interaction in self.model.yield_interactions():

                # regulatory event over interaction.regulatory events.
                # an interaction can have multiple regulatory events.
                # Regularly, there is one event for each possible coefficient of the target variable.
                # If the regulatory event is evaluated to True or 1, this is the outcome coefficient of the target
                # By default the outcome of the target is a zero coefficient

                is_empty = True
                active_coefficient = None

                for coefficient, regulatory_event in interaction.regulatory_events.items():

                    if not regulatory_event.is_none:

                        is_empty = False

                        regulatory_event_eval = regulatory_event.evaluate(values=state)

                        if regulatory_event_eval:
                            active_coefficient = coefficient

                            break

                if is_empty:
                    target_val = interaction.target.coefficient.active_coefficient

                elif active_coefficient is not None:
                    target_val = active_coefficient

                else:
                    target_val = 0.0

                result[interaction.target.id] = target_val

        return result

    def decode_metabolic_state(self, state):
        """
        Method responsible for decoding the RFBA metabolic state, namely the state of all metabolic genes associated
        at least with one reaction in the GPRs rule.

        :param state: dict of regulatory/metabolic variable keys (metabolic target) and a value of 0 or 1

        :return: dict of constraints of the resulting metabolic state
        """

        # This method retrieves the constraints associated with a given metabolic/regulatory state

        constraints = {}

        if self.model.is_metabolic():

            # gpr over model.reactions
            for rxn in self.model.yield_reactions():

                if rxn.gpr.is_none:
                    continue

                else:

                    res = rxn.gpr.evaluate(values=state)

                    if not res:
                        constraints[rxn.id] = (0.0, 0.0)

        return constraints

    def next_state(self, state, constraints):
        """
        Retrieves the next state for a given state

        Solves the boolean regulatory model using method regulatory_simulation(state) and decodes the metabolic state
        for that state or initial state (according to the flag regulatory_state)

        :param state: dict of regulatory/metabolic variable keys (regulatory and metabolic target) and a value of 0, 1
        or float (reactions and metabolites predicates)
        :param constraints:

        :return: dict of all regulatory/metabolic variables keys and a value of the resulting state
        """

        # Regulatory rules from the boolean network
        regulatory_state = self.sim_bool(state=state)

        # Reactions and metabolites inactivated by the genes, reactions and metabolites
        # in the initial state or regulatory state simulation
        metabolic_state = {gene.id: gene.coefficient.active_coefficient
                           for gene in self.model.yield_genes()}
        metabolic_regulatory_state = {**metabolic_state, **regulatory_state}

        regulatory_constraints = self.decode_metabolic_state(metabolic_regulatory_state)
        metabolic_regulatory_constraints = {**constraints, **regulatory_constraints}

        solver_solution = self._solver.solve(linear=None,
                                             quadratic=None,
                                             minimize=None,
                                             model=None,
                                             constraints=metabolic_regulatory_constraints,
                                             get_values=True,
                                             shadow_prices=False,
                                             reduced_costs=False,
                                             pool_size=0,
                                             pool_gap=None)

        if not solver_solution.values:
            solver_solution.values = {}

        next_state = {**state, **metabolic_regulatory_state}

        # update the regulatory/metabolic state
        for reaction in self._regulatory_reactions:

            next_state[reaction] = solver_solution.values.get(reaction, 0.0)

        for metabolite in self._regulatory_metabolites:

            reaction = self.model.get(metabolite).exchange_reaction

            if reaction:

                next_state[metabolite] = solver_solution.values.get(reaction.id, 0.0)

        return next_state

    def steady_state_optimize(self,
                              initial_state=None,
                              objective=None,
                              minimize=False,
                              constraints=None,
                              to_solver=False,
                              get_values=True,
                              shadow_prices=False,
                              reduced_costs=False,
                              pool_size=0,
                              pool_gap=None):
        """
        RFBA model one-step simulation (pseudo steady-state).

        First, the boolean regulatory model is simulated using the initial state. If there is no initial state ,
        the state of all regulatory variables is inferred from the model.

        At the same time, the metabolic state is also decoded from the initial state of all metabolic genes.
        If otherwise set by using the regulatory flag, the metabolic state is decoded from the boolean regulatory model
        simulation.

        The result of the boolean regulatory model updates the state of all targets, while the result of the
        metabolic state decoding updates the state of all reactions and metabolites that are also regulatory variables.

        Then, this new state is used to decode the final metabolic state using the resulting metabolic-regulatory state.

        Objective and constraint-based model analysis method (pFBA, FBA, ...) can be altered here reversibly.

        :return: ModelSolution object with metabolic/regulatory solution
        """

        self.set_objective(linear=objective, minimize=minimize)

        # It takes the initial state from the model and then updates with the initial state provide as input
        if not constraints:
            constraints = {}
        else:
            constraints = constraints.copy()

        if not initial_state:
            initial_state = self.initial_state()

        else:
            initial_state = {**self.initial_state(), **initial_state}

        # Regulatory rules from the boolean network
        regulatory_state = self.sim_bool(state=initial_state)

        # Reactions and metabolites inactivated by the genes, reactions and metabolites
        # in the initial state or regulatory state simulation
        metabolic_state = {gene.id: gene.coefficient.active_coefficient
                           for gene in self.model.yield_genes()}
        metabolic_regulatory_state = {**metabolic_state, **regulatory_state}

        regulatory_constraints = self.decode_metabolic_state(metabolic_regulatory_state)
        metabolic_regulatory_constraints = {**constraints, **regulatory_constraints}

        solver_solution = self._solver.solve(linear=None,
                                             quadratic=None,
                                             minimize=None,
                                             model=None,
                                             constraints=metabolic_regulatory_constraints,
                                             get_values=get_values,
                                             shadow_prices=shadow_prices,
                                             reduced_costs=reduced_costs,
                                             pool_size=pool_size,
                                             pool_gap=pool_gap)

        if solver_solution.values:
            solver_solution.values = {**initial_state, **metabolic_regulatory_state, **solver_solution.values}

        if to_solver:
            return solver_solution

        if minimize:
            sense = 'minimize'

        else:
            sense = 'maximize'

        return ModelSolution.from_solver(method=self.method,
                                         solution=solver_solution,
                                         model=self.model,
                                         objective_direction=sense)

    def dynamic_optimize(self,
                         initial_state=None,
                         iter_limit=10,
                         objective=None,
                         minimize=False,
                         constraints=None,
                         to_solver=False,
                         get_values=True,
                         shadow_prices=False,
                         reduced_costs=False,
                         pool_size=0,
                         pool_gap=None):
        """
        RFBA model dynamic simulation (until the metabolic-regulatory state is reached).

        First, the boolean regulatory model is solved using the initial state. If there is no initial state ,
        the state of all regulatory variables is inferred from the model.

        At the same time, the metabolic state is also decoded from the initial state of all metabolic genes.
        If otherwise set by using the regulatory flag, the metabolic state is decoded from the boolean regulatory model
        simulation.

        The result of the boolean regulatory model updates the state of all targets, while the result of the
        metabolic state decoding updates the state of all reactions and metabolites that are also regulatory variables.

        Then, this new resulting metabolic-regulatory state is used to solve again the boolean regulatory model and
        decode again the metabolic state towards the retrieval of a new regulatory and metabolic states.

        This cycle is iterated until a given state is repeated, the so called metabolic-regulatory steady-state.
        Hence, dynamic RFBA is based on finding the cyclic attractors of the metabolic-regulatory networks
        Alternatively, an iteration limit may be reached.

        Finally, all states between the cyclic attractor are used for decoding final metabolic states using
        the resulting metabolic-regulatory state. Thus, a solution/simulation is obtained for each mid-state in the
        cyclic attractor of the metabolic-regulatory networks

        Objective and constraint-based model analysis method (pFBA, FBA, ...) can be altered here reversibly.

        :return:
        """

        self.set_objective(linear=objective, minimize=minimize)

        # It takes the initial state from the model and then updates with the initial state provide as input
        if not constraints:
            constraints = {}
        else:
            constraints = constraints.copy()

        if not initial_state:
            initial_state = self.initial_state()

        else:
            initial_state = {**self.initial_state(), **initial_state}

        regulatory_solution = []

        # solve using the initial state
        state = self.next_state(state=initial_state, constraints=constraints)
        regulatory_solution.append(state)

        i = 1
        sol = 1
        steady_state = False
        while not steady_state:

            # Updating state upon state. See next state for further detail
            state = self.next_state(state=state, constraints=constraints)
            regulatory_solution.append(state)

            n_sols = len(regulatory_solution) - 1

            for sol in range(n_sols):

                # A reaction can have different flux values between two solutions, though the state remains the same.
                # Thus, non-zero flux stands for ON state while zero flux stands for OFF state

                replica_v = []

                for key, val in regulatory_solution[sol].items():

                    sol_val = regulatory_solution[n_sols][key]

                    if key in self._regulatory_reactions or key in self._regulatory_metabolites:

                        if abs(val) > ModelConstants.TOLERANCE:
                            val = True
                        else:
                            val = False

                        if abs(sol_val) > ModelConstants.TOLERANCE:
                            sol_val = True
                        else:
                            sol_val = False

                    replica_v.append(val == sol_val)

                is_replica = all(replica_v)

                if is_replica:
                    steady_state = True

                    break

            if i < iter_limit:
                i += 1
            else:

                def iteration_limit(message):
                    warn(message, UserWarning, stacklevel=2)

                iteration_limit("Iteration limit reached")
                steady_state = True

        # The solution comprehends and FBA simulation between the first and second similar regulatory and metabolic
        # states using the corresponding regulatory states to switch on or off the associated reactions
        solution = []
        for i in range(sol, len(regulatory_solution)):

            regulatory_state = regulatory_solution[i]

            regulatory_state_constraints = self.decode_metabolic_state(regulatory_state)

            state_constraints = {**constraints, **regulatory_state_constraints}

            solver_solution = self._solver.solve(linear=None,
                                                 quadratic=None,
                                                 minimize=None,
                                                 model=None,
                                                 constraints=state_constraints,
                                                 get_values=get_values,
                                                 shadow_prices=shadow_prices,
                                                 reduced_costs=reduced_costs,
                                                 pool_size=pool_size,
                                                 pool_gap=pool_gap)

            if solver_solution.values:
                solver_solution.values = {**regulatory_state, **solver_solution.values}

            if to_solver:
                solution.append(solver_solution)

                continue

            if minimize:
                sense = 'minimize'

            else:
                sense = 'maximize'

            sol = ModelSolution.from_solver(method=self.method,
                                            solution=solver_solution,
                                            model=self.model,
                                            objective_direction=sense)

            solution.append(sol)

        if to_solver:
            return solution

        return DynamicSolution(*solution)

    def optimize(self,
                 initial_state=None,
                 dynamic=False,
                 iter_limit=10,
                 objective=None,
                 minimize=False,
                 constraints=None,
                 to_solver=False,
                 get_values=True,
                 shadow_prices=False,
                 reduced_costs=False,
                 pool_size=0,
                 pool_gap=None):

        if dynamic:
            return self.dynamic_optimize(initial_state=initial_state,
                                         iter_limit=iter_limit,
                                         objective=objective,
                                         minimize=minimize,
                                         constraints=constraints,
                                         to_solver=to_solver,
                                         get_values=get_values,
                                         shadow_prices=shadow_prices,
                                         reduced_costs=reduced_costs,
                                         pool_size=pool_size,
                                         pool_gap=pool_gap)

        return self.steady_state_optimize(initial_state=initial_state,
                                          objective=objective,
                                          minimize=minimize,
                                          constraints=constraints,
                                          to_solver=to_solver,
                                          get_values=get_values,
                                          shadow_prices=shadow_prices,
                                          reduced_costs=reduced_costs,
                                          pool_size=pool_size,
                                          pool_gap=pool_gap)



from typing import Union

from mewpy.solvers.solver import Solver
from mewpy.mew.solution import ModelSolution
from mewpy.model import Model, MetabolicModel, RegulatoryModel


# TODO: type hinting and documentation
class SimBool:

    def __init__(self,
                 model: Union[Model, MetabolicModel, RegulatoryModel],
                 solver: Union[str, Solver, None] = None,
                 build: bool = True,
                 attach: bool = True):
        """
        Simulation of a Boolean network (SimBool) for a regulatory model.

        :param model: a mewpy Model, MetabolicModel, RegulatoryModel or all. The model is used to retrieve
        variables and constraints to the linear problem

        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used but it will be overwritten
        if build is true.

        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False
        """

        self.model = model

    def build(self):

        pass

    def optimize(self, initial_state=None, to_dict=False):

        res = {}

        # solving regulatory rule by regulatory rule (synchronously though, as asynchronously would take too much time)
        # Each target has one and only one regulatory interaction
        # The regulatory model object only supports this strategy, though this can easily be altered. If so,
        # this solve method must be changed too

        if not initial_state:
            initial_state = {}

        else:
            initial_state = initial_state.copy()

        if self.model.is_regulatory():

            reg_state = {reg.id: reg.coefficient.active_coefficient
                         for reg in self.model.yield_regulators()}

            # update with the most recent state
            reg_state.update(initial_state)

            # interaction over model.interactions.

            for interaction in self.model.yield_interactions():

                # regulatory event over interaction.regulatory events.
                # an interaction can have multiple regulatory events.
                # Regularly, there is one event for each possible coefficient of the target variable.
                # If the regulatory event is evaluated to True or 1, this is the outcome coefficient of the target
                # By default the outcome of the target is a zero coefficient

                target_val = 0

                for coefficient, regulatory_event in interaction.regulatory_events.items():

                    reg_event_val = regulatory_event.evaluate(values=reg_state)

                    if reg_event_val:
                        target_val = coefficient

                res[interaction.target.id] = target_val

        if to_dict:
            return res

        return ModelSolution(method='SynchronousBoolean', x=res, objective_value=0.0, status='Feasible')



# TODO: this module largely depends on pandas dataframes. Should it be set as package requirement?
# noinspection PyPackageRequirements
from pandas import concat


# TODO: methods stubs and type hinting
# TODO: maybe the following solution objects should inherit from the same multisolution object!
class MultiSolution:

    def __init__(self, *solutions):

        if not solutions:
            _solutions = {}
        else:
            _solutions = {}

            for solution in solutions:
                setattr(self, f'{solution.method}', solution)
                _solutions[solution.method] = solution

        self._solutions = _solutions

    @property
    def solutions(self):
        return self._solutions

    def to_frame(self):
        frames = []
        columns = []

        for method, solution in self._solutions.items():
            frames.append(solution.to_frame().frame)
            columns.append(method)

        df = concat(frames, axis=1, join='outer', keys=columns)

        return df

    def to_summary(self):

        frames = []
        columns = []

        for method, solution in self._solutions.items():
            frames.append(solution.to_summary().frame)
            columns.append(method)

        df = concat(frames, axis=1, join='outer', keys=columns)

        return df

    def __repr__(self):
        return 'MultiSolution'

    def __str__(self):
        return 'MultiSolution:' ','.join([method for method in self._solutions])


# TODO: methods stubs and type hinting
class DynamicSolution:

    def __init__(self, *solutions, time=None):

        if not solutions:
            _solutions = {}
            time = []

        else:
            _solutions = {}

            if time is None:
                time = [i for i in range(len(solutions))]

            else:
                time = list(time)

            for t, solution in zip(time, solutions):
                setattr(self, f't_{t}', solution)
                _solutions[f't_{t}'] = solution

        self._solutions = _solutions
        self._time = time

    @property
    def solutions(self):
        return self._solutions

    def to_frame(self):
        frames = []
        columns = []

        for time, solution in self._solutions.items():
            frames.append(solution.to_frame().frame)
            columns.append(time)

        df = concat(frames, axis=1, join='outer', keys=columns)

        return df

    def to_summary(self):

        frames = []
        columns = []

        for time, solution in self._solutions.items():
            frames.append(solution.to_summary().frame)
            columns.append(time)

        df = concat(frames, axis=1, join='outer', keys=columns)

        return df

    def __repr__(self):
        return 'DynamicSolution'

    def __str__(self):
        return 'DynamicSolution:' ','.join([time for time in self._solutions])


# TODO: methods stubs and type hinting
class KOSolution:

    def __init__(self, solutions, kos=None):

        if not solutions:
            _solutions = {}
            kos = []

        else:
            _solutions = {}

            if not kos:

                if isinstance(solutions, dict):

                    kos = list(solutions.keys())
                    solutions = list(solutions.values())

                else:

                    kos = [i for i, _ in enumerate(solutions)]

            else:

                if isinstance(solutions, dict):
                    solutions = list(solutions.values())

            for ko, solution in zip(kos, solutions):
                setattr(self, f'ko_{ko}', solution)
                _solutions[f'ko_{ko}'] = solution

        self._solutions = _solutions
        self._time = kos

    @property
    def solutions(self):
        return self._solutions

    def to_frame(self):
        frames = []
        columns = []

        for ko, solution in self._solutions.items():
            frames.append(solution.to_frame().frame)
            columns.append(ko)

        df = concat(frames, axis=1, join='outer', keys=columns)

        return df

    def to_summary(self):

        frames = []
        columns = []

        for ko, solution in self._solutions.items():
            frames.append(solution.to_summary().frame)
            columns.append(ko)

        df = concat(frames, axis=1, join='outer', keys=columns)

        return df

    def __repr__(self):
        return 'KOSolution'

    def __str__(self):
        return 'KOSolution:' ','.join([time for time in self._solutions])



from typing import Any, Dict, Union, TYPE_CHECKING, Tuple, Generator, List

from mewpy.mew.algebra import Expression, parse_expression
from mewpy.mew.lp import Notification
from mewpy.util.utilities import generator
from mewpy.util.serialization import serialize
from mewpy.util.history import recorder
from mewpy.util.constants import ModelConstants
from mewpy.io.engines.engines_utils import expression_warning
from .coefficient import Coefficient
from .variable import Variable, variables_from_symbolic

if TYPE_CHECKING:
    from .metabolite import Metabolite
    from .gene import Gene
    from mewpy.model import Model, MetabolicModel, RegulatoryModel


# TODO: methods stubs
class Reaction(Variable, variable_type='reaction', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 bounds: Tuple[Union[float, int], Union[float, int]] = None,
                 stoichiometry: Dict['Metabolite', Union[float, int]] = None,
                 gpr: Expression = None,
                 **kwargs):

        """
        A set of reactions is the basis for all metabolic models.
        This reaction object holds information for bounds, metabolites, stoichiometry and
        gene-protein-reaction rules (the associated between reactions, enzymes/proteins and genes)

        A reaction usually contains the stoichiometry of the metabolites that take part in the reaction.
        A reaction usually contains the GPR rule/logic of the associated genes
        A reaction usually contains the bounds for the fluxes values that can take during simulation

        Dynamic information can be obtained from the previous attributes, such as metabolites, reactants, products,
        genes, boundary, reversibility, charge and mass balance, equation, compartments, etc.

        A reaction is the main object for metabolic models and thus the object that offers more manipulation solutions.

        :param identifier: identifier, e.g. PYK
        :param bounds: tuple having both lower and upper bounds, respectively the minimum and maximum coefficient
        that a reaction can have
        :param stoichiometry: dictionary of metabolites objects and their corresponding stoichiometric value
        :param gpr: an expression object with the boolean logic of the associated genes.
        Genes objects must be associated with the expression object.
        """

        # the coefficient bounds initializer sets minimum and maximum bounds of MEWPY_LB and MEWPY_UB
        if not bounds:
            lb, ub = ModelConstants.REACTION_LOWER_BOUND, ModelConstants.REACTION_UPPER_BOUND

        else:
            lb, ub = bounds

        if not gpr:
            gpr = Expression()

        if not stoichiometry:
            stoichiometry = {}

        self._bounds = Coefficient.from_bounds(variable=self, lb=lb, ub=ub)
        self._gpr = Expression()
        self._stoichiometry = {}

        super().__init__(identifier,
                         **kwargs)

        self.replace_stoichiometry(stoichiometry, history=False)

        # The gpr must be an expression object. The expression object properties are used to check whether the genes
        # container is correct.
        self.add_gpr(gpr, history=False)

        # There is an alternative constructor for building a reaction with a stringify-like gpr rule.
        # Parsing takes the hard job to infer the correct relation between genes

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------

    @property
    def types(self):

        # noinspection PyUnresolvedReferences
        _types = {Reaction.variable_type}

        _types.update(super(Reaction, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------

    def __str__(self):
        return f'{self.id}: {self.equation}'

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('stoichiometry', 'stoichiometry', '_stoichiometry')
    @property
    def stoichiometry(self) -> Dict['Metabolite', Union[int, float]]:
        return self._stoichiometry.copy()

    @serialize('coefficient', 'bounds', '_bounds')
    @property
    def coefficient(self) -> Coefficient:
        return self._bounds

    @serialize('gpr', 'gpr', '_gpr')
    @property
    def gpr(self) -> Expression:
        return self._gpr

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------

    @stoichiometry.setter
    @recorder
    def stoichiometry(self, value: Dict['Metabolite', Union[int, float]]):

        self.replace_stoichiometry(value, history=False)

    @gpr.setter
    @recorder
    def gpr(self, value: Expression):

        self.remove_gpr(history=False)
        self.add_gpr(value, history=False)

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------

    @property
    def metabolites(self) -> Dict[str, 'Metabolite']:
        return {met.id: met for met in self._stoichiometry}

    @property
    def products(self) -> Dict[str, 'Metabolite']:
        return {met.id: met for met, st in self._stoichiometry.items()
                if st > 0.0}

    @property
    def reactants(self) -> Dict[str, 'Metabolite']:
        return {met.id: met for met, st in self._stoichiometry.items()
                if st < 0.0}

    @property
    def compartments(self) -> set:

        return {met.compartment
                for met in self.yield_metabolites()
                if met.compartment is not None}

    @property
    def bounds(self) -> Tuple[Union[int, float], Union[int, float]]:
        return self.coefficient.bounds

    @property
    def lower_bound(self) -> Union[int, float]:
        return self.coefficient.lower_bound

    @property
    def upper_bound(self) -> Union[int, float]:
        return self.coefficient.upper_bound

    @property
    def reversibility(self) -> bool:
        return self.lower_bound < - ModelConstants.TOLERANCE and self.upper_bound > ModelConstants.TOLERANCE

    @property
    def equation(self) -> str:

        equation = ''

        for _id, met in self.reactants.items():
            equation += f'{-self._stoichiometry[met]} {_id} + '

        if self.reversibility:
            equation = equation[:-2] + '<-> '
        else:
            equation = equation[:-2] + '-> '

        for _id, met in self.products.items():
            equation += f'{self._stoichiometry[met]} {_id} + '

        if not equation:
            return ''

        elif self.products:
            return equation[:-2]

        else:
            return equation[:-1]

    @property
    def genes(self) -> Dict[str, 'Gene']:

        return self.gpr.variables.copy()

    @property
    def gene_protein_reaction_rule(self) -> str:
        return self.gpr.to_string()

    @property
    def boundary(self) -> bool:
        return not (self.reactants and self.products)

    @property
    def charge_balance(self) -> Dict[str, Union[int, float]]:
        return {'reactants': sum([self._stoichiometry[met] * met.charge for met in self.yield_reactants()]),
                'products': sum([self._stoichiometry[met] * met.charge for met in self.yield_products()])}

    @property
    def mass_balance(self) -> Dict[str, Union[int, float]]:

        balance = {}

        for metabolite, st in self._stoichiometry.items():

            for atom, count in metabolite.atoms.items():
                amount = count * st

                balance[atom] = balance.get(atom, 0) + amount

        return balance

    # -----------------------------------------------------------------------------
    # Dynamic attributes setters
    # -----------------------------------------------------------------------------

    @bounds.setter
    @recorder
    def bounds(self, value: Tuple[Union[float, int], Union[float, int]]):

        if not value:
            value = (ModelConstants.REACTION_LOWER_BOUND, ModelConstants.REACTION_UPPER_BOUND)

        self.coefficient.coefficients = value

    @lower_bound.setter
    @recorder
    def lower_bound(self, value: Union[float, int]):

        if not value:
            value = ModelConstants.REACTION_LOWER_BOUND

        value = (value,  self.upper_bound)

        self.coefficient.coefficients = value

    @upper_bound.setter
    @recorder
    def upper_bound(self, value: Union[float, int]):

        if not value:
            value = ModelConstants.REACTION_UPPER_BOUND

        value = (self.lower_bound, value)

        self.coefficient.coefficients = value

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------

    def yield_genes(self) -> Generator['Gene', None, None]:

        return generator(self.genes)

    def yield_metabolites(self) -> Generator['Metabolite', None, None]:

        return generator(self.metabolites)

    def yield_reactants(self) -> Generator['Metabolite', None, None]:

        return generator(self.reactants)

    def yield_products(self) -> Generator['Metabolite', None, None]:

        return generator(self.products)

    # -----------------------------------------------------------------------------
    # Polymorphic constructors
    # -----------------------------------------------------------------------------

    @classmethod
    def from_stoichiometry(cls,
                           stoichiometry: Dict['Metabolite', Union[float, int]],
                           identifier: Any,
                           bounds: Tuple[Union[float, int], Union[float, int]] = None,
                           gpr: Expression = None,
                           model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                           **kwargs) -> 'Reaction':

        instance = cls(identifier=identifier,
                       bounds=bounds,
                       stoichiometry=stoichiometry,
                       gpr=gpr,
                       model=model,
                       **kwargs)

        return instance

    @classmethod
    def from_gpr_string(cls,
                        gpr_string: str,
                        identifier: Any,
                        bounds: Tuple[Union[float, int], Union[float, int]] = None,
                        stoichiometry: Dict['Metabolite', Union[float, int]] = None,
                        model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                        **kwargs) -> 'Reaction':

        try:

            symbolic = parse_expression(gpr_string)

            genes = variables_from_symbolic(symbolic=symbolic, types=('gene', ), model=model)

            expression = Expression(symbolic=symbolic, variables=genes)

        except SyntaxError as exc:

            expression_warning(f'{gpr_string} cannot be parsed')

            raise exc

        instance = cls(identifier=identifier,
                       bounds=bounds,
                       stoichiometry=stoichiometry,
                       gpr=expression,
                       model=model,
                       **kwargs)

        return instance

    @classmethod
    def from_gpr_expression(cls,
                            gpr: Expression,
                            identifier: Any,
                            bounds: Tuple[Union[float, int], Union[float, int]] = None,
                            stoichiometry: Dict['Metabolite', Union[float, int]] = None,
                            model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                            **kwargs) -> 'Reaction':

        if not isinstance(gpr, Expression):
            raise TypeError(f'expression must be an {Expression} object')

        instance = cls(identifier=identifier,
                       bounds=bounds,
                       stoichiometry=stoichiometry,
                       gpr=gpr,
                       model=model,
                       **kwargs)

        return instance

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    def ko(self, minimum_coefficient: Union[int, float] = 0.0, history=True):

        return self.coefficient.ko(minimum_coefficient=minimum_coefficient, history=history)

    def update(self,
               bounds: Tuple[Union[float, int], Union[float, int]] = None,
               stoichiometry: Dict['Metabolite', Union[float, int]] = None,
               gpr: Expression = None,
               **kwargs):

        super(Reaction, self).update(**kwargs)

        if bounds is not None:
            self.bounds = bounds

        if gpr is not None:
            self.gpr = gpr

        if stoichiometry is not None:
            self.stoichiometry = stoichiometry

    def replace_stoichiometry(self,
                              stoichiometry: Dict['Metabolite', Union[float, int]],
                              remove_orphans_from_model: bool = True,
                              history=True):

        if history:
            self.history.queue_command(undo_func=self.replace_stoichiometry,
                                       undo_kwargs={'stoichiometry': self.stoichiometry,
                                                    'remove_orphans_from_model': remove_orphans_from_model,
                                                    'history': False},
                                       func=self.replace_stoichiometry,
                                       kwargs={'stoichiometry': stoichiometry,
                                               'remove_orphans_from_model': remove_orphans_from_model,
                                               'history': history})

        if not stoichiometry and not self._stoichiometry:
            self._stoichiometry = {}

            return

        self.remove_metabolites(self.yield_metabolites(),
                                remove_orphans_from_model=remove_orphans_from_model,
                                history=False)

        self.add_metabolites(stoichiometry,
                             history=False)

    def add_metabolites(self,
                        stoichiometry: Dict['Metabolite', Union[float, int]],
                        history=True):

        to_add = []

        for metabolite, coefficient in stoichiometry.items():

            # noinspection PyProtectedMember
            metabolite._reactions[self.id] = self

            self._stoichiometry[metabolite] = coefficient

            to_add.append(metabolite)

        if self.model:
            # the add interface will add all metabolites to the simulators. The simulators will retrieve the new
            # constraints for each metabolite and replace it in the LP
            self.model.add(to_add, 'metabolite', comprehensive=False, history=False)

            notification = Notification(content=(self, ),
                                        content_type='reactions',
                                        action='add')

            self.model.notify(notification)

        if history:
            self.history.queue_command(undo_func=self.remove_metabolites,
                                       undo_kwargs={'metabolites': list(stoichiometry.keys()),
                                                    'remove_orphans_from_model': True,
                                                    'history': False},
                                       func=self.add_metabolites,
                                       kwargs={'stoichiometry': stoichiometry,
                                               'history': history})

    def remove_metabolites(self,
                           metabolites: List['Metabolite'],
                           remove_orphans_from_model: bool = True,
                           history=True):

        if isinstance(metabolites, dict):
            metabolites = list(metabolites.values())

        orphan_mets = []
        old_stoichiometry = {}

        for metabolite in metabolites:

            # noinspection PyProtectedMember
            del metabolite._reactions[self.id]

            # noinspection PyProtectedMember
            if not metabolite._reactions:
                orphan_mets.append(metabolite)

            old_stoichiometry[metabolite] = self._stoichiometry[metabolite]

            del self._stoichiometry[metabolite]

        if self.model:

            # the add interface will add all metabolites to the simulators. The simulators will retrieve the new
            # constraints for each metabolite and replace it in the LP

            if orphan_mets and remove_orphans_from_model:
                self.model.remove(orphan_mets, 'metabolite', remove_orphans=False, history=False)

            notification = Notification(content=(self, ),
                                        content_type='reactions',
                                        action='add')

            self.model.notify(notification)

        if history:
            self.history.queue_command(undo_func=self.add_metabolites,
                                       undo_kwargs={'stoichiometry': old_stoichiometry,
                                                    'history': False},
                                       func=self.remove_metabolites,
                                       kwargs={'metabolites': metabolites,
                                               'remove_orphans_from_model': remove_orphans_from_model,
                                               'history': history})

    def add_gpr(self, gpr: Expression, history=True):

        if not isinstance(gpr, Expression):
            raise TypeError(f'expression must be an {Expression} object. '
                            f'To set None, provide an empty Expression()')

        if history:
            self.history.queue_command(undo_func=self.remove_gpr,
                                       undo_kwargs={'remove_orphans': True,
                                                    'history': False},
                                       func=self.add_gpr,
                                       kwargs={'gpr': gpr,
                                               'history': history})

        if self.gpr:
            self.remove_gpr(history=False)

        to_add = []

        for gene in gpr.variables.values():
            gene.update(reactions={self.id: self},
                        model=self.model)

            to_add.append(gene)

        self._gpr = gpr

        if self.model:
            self.model.add(to_add, 'gene', comprehensive=False, history=False)

            notification = Notification(content=(self,),
                                        content_type='gprs',
                                        action='add')

            self.model.notify(notification)

    def remove_gpr(self, remove_orphans: bool = True, history=True):

        if history:
            self.history.queue_command(undo_func=self.add_gpr,
                                       undo_kwargs={'gpr': self.gpr,
                                                    'history': False},
                                       func=self.remove_gpr,
                                       kwargs={'remove_orphans': remove_orphans,
                                               'history': history})

        to_remove = []

        for gene in self.yield_genes():

            # noinspection PyProtectedMember
            del gene._reactions[self.id]

            # noinspection PyProtectedMember
            if not gene._reactions:
                to_remove.append(gene)

        self._gpr = Expression()

        if self.model:

            if remove_orphans:
                self.model.remove(to_remove, 'gene', remove_orphans=False, history=False)

            notification = Notification(content=(self,),
                                        content_type='gprs',
                                        action='add')

            self.model.notify(notification)



from enum import Enum

from ..util.constants import ModelConstants

solvers = []

try:
    import gurobipy
    solvers.append('gurobi')
except ImportError as e:
    pass

if 'gurobi' in solvers:
    try:
        # disables solver output.
        gurobipy.setParam('OutputFlag', 0)
    except:
        pass

try:
    import cplex
    solvers.append('cplex')
except ImportError as e:
    pass

try:
    import swiglpk
    solvers.append('glpk')
except ImportError:
    pass

default_solver = None


def get_default_solver():
    """
    Returns:
        [type]: [description]
    """
    global default_solver

    if default_solver:
        return default_solver

    solver_order = ['cplex', 'gurobi', 'glpk']

    for solver in solver_order:
        if solver in solvers:
            default_solver = solver
            break

    if not default_solver:
        raise RuntimeError("No solver available.")

    return default_solver


def set_default_solver(solvername):
    """ Sets default solver.
    Arguments:
        solvername : (str) solver name (currently available: 'gurobi', 'cplex')
    """

    global default_solver

    if solvername.lower() in solvers:
        default_solver = solvername.lower()
    else:
        raise RuntimeError(f"Solver {solvername} not available.")


# Model specific simulators mapping:
# Entries take the form:  full_model_class_path -> (simulator_path, simulator_class_name)
# TODO: use qualified names
map_model_simulator = {
    'geckopy.gecko.GeckoModel': ('mewpy.simulation.cobra', 'GeckoSimulation'),
    'mewpy.model.gecko.GeckoModel': ('mewpy.simulation.reframed', 'GeckoSimulation'),
    'mewpy.model.smoment.SMomentModel': ('mewpy.simulation.reframed', 'GeckoSimulation'),
    'mewpy.model.model.Model': ('mewpy.simulation.mew', 'Simulation'),
    'mewpy.model.metabolic.MetabolicModel': ('mewpy.simulation.mew', 'Simulation'),
    'mewpy.model.regulatory.RegulatoryModel': ('mewpy.simulation.mew', 'Simulation'),
    'mewpy.model.model.MetabolicRegulatoryModel': ('mewpy.simulation.mew', 'Simulation'),
    'mewpy.model.model.RegulatoryMetabolicModel': ('mewpy.simulation.mew', 'Simulation'),
}


def get_simulator(model, envcond=None, constraints=None, reference=None, reset_solver=ModelConstants.RESET_SOLVER):
    """
    Returns a simulator instance for the model.
    The simulator instance is model dependent.
    Besides able to be used on its own, this function is invoked by EA optimization problems
    and by evaluation function instances to perform phenotyoe evaluations of candidate solutions.


    :param model: the model
    :param dict envcond: A dictionary of environmental conditions.
    :param dict constraints: A dictionary of additional persistent constraints.
    :param dict reference: A dictionary of the wild type flux values
    :param bool reset_solver: Whether to reset the solver before each simulation
    :returns: An instance of Simulator
    """

    instance = None
    name = f"{model.__class__.__module__}.{model.__class__.__name__}"
    if name in map_model_simulator:
        module_name, class_name = map_model_simulator[name]
        module = __import__(module_name, fromlist=[None])
        class_ = getattr(module, class_name)
        try:
            model.solver.configuration.timeout = ModelConstants.SOLVER_TIMEOUT
        except:
            pass
        instance = class_(model, envcond=envcond,
                          constraints=constraints, reference=reference, reset_solver=reset_solver)
    elif "etfl" in name:
        try:
            from .cobra import Simulation
            from etfl.optim.config import standard_solver_config
            standard_solver_config(model, verbose=False)
            model.solver.configuration.timeout = max(7200, ModelConstants.SOLVER_TIMEOUT)
            instance = Simulation(
                model, envcond=envcond, constraints=constraints, reference=reference, reset_solver=reset_solver)
            instance._MAX_STR = 'max'
            instance._MIN_STR = 'min'
        except Exception:
            raise RuntimeError("Could not create simulator for the ETFL model")
    else:
        try:
            from cobra.core.model import Model
            if isinstance(model, Model):
                from .cobra import Simulation
                model.solver.configuration.timeout = ModelConstants.SOLVER_TIMEOUT
                instance = Simulation(
                    model, envcond=envcond, constraints=constraints, reference=reference, reset_solver=reset_solver)
        except ImportError:
            pass
        if not instance:
            try:
                from reframed.core.cbmodel import CBModel
                if isinstance(model, CBModel):
                    from .reframed import Simulation
                    instance = Simulation(
                        model, envcond=envcond, constraints=constraints, reference=reference, reset_solver=reset_solver)
            except ImportError:
                pass

    if not instance:
        raise ValueError(f"The model <{name}> has no defined simulator.")
    return instance


def get_container(model):
    """
    Returns a container for a given model sharing a common interface.
    A container does not perform any task, it only serves as a basic interface with a phenotype simulator.

    :param model: An instance of a metabolic model.
    :returns: A container.

    """

    from mewpy.model import Model, RegulatoryModel, MetabolicModel

    if isinstance(model, (Model, MetabolicModel, RegulatoryModel)):

        from .mew import MewModelContainer

        return MewModelContainer(model)

    try:
        from reframed.core.cbmodel import CBModel
        if isinstance(model, CBModel):
            from mewpy.simulation.reframed import CBModelContainer
            return CBModelContainer(model)
    except Exception:
        pass

    try:
        from cobra.core.model import Model
        if isinstance(model, Model):
            from mewpy.simulation.cobra import CobraModelContainer
            return CobraModelContainer(model)
    except Exception:
        pass

    raise ValueError(f"Unrecognized model class: {model.__class__.name}")


class SimulationMethod(Enum):
    FBA = 'FBA'
    pFBA = 'pFBA'
    MOMA = 'MOMA'
    lMOMA = 'lMOMA'
    ROOM = 'ROOM'
    NONE = 'NONE'

    def __eq__(self, other):
        """Overrides equal to enable string name comparison.
        Allows to seamlessly use:
            SimulationMethod.FBA = SimulationMethod.FBA
            SimulationMethod.FBA = 'FBA'
        without requiring an additional level of comparison (SimulationMethod.FBA.name = 'FBA')
        """
        if isinstance(other, SimulationMethod):
            return super().__eq__(other)
        elif isinstance(other, str):
            return self.name == other
        else:
            return False

    def __hash__(self):
        return hash(self.name)


class SStatus(Enum):
    """ Enumeration of possible solution status. """
    OPTIMAL = 'Optimal'
    UNKNOWN = 'Unknown'
    SUBOPTIMAL = 'Suboptimal'
    UNBOUNDED = 'Unbounded'
    INFEASIBLE = 'Infeasible'
    INF_OR_UNB = 'Infeasible or Unbounded'

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name
